import json
from unittest.mock import MagicMock

from pypanther import LogType, Rule, RuleMock, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get
from pypanther.helpers.panther_base_helpers import gsuite_parameter_lookup as param_lookup

g_suite_drive_visibility_changed_tests: list[RuleTest] = [
    RuleTest(
        name="Access Event",
        expected_result=False,
        log={
            "p_row_id": "111222",
            "actor": {"email": "bobert@example.com"},
            "id": {"applicationName": "drive"},
            "events": [{"type": "access", "name": "upload"}],
        },
    ),
    RuleTest(
        name="ACL Change without Visibility Change",
        expected_result=False,
        log={
            "p_row_id": "111222",
            "actor": {"email": "bobert@example.com"},
            "id": {"applicationName": "drive"},
            "events": [{"type": "acl_change", "name": "shared_drive_settings_change"}],
        },
    ),
    RuleTest(
        name="Doc Became Public - Link (Unrestricted)",
        expected_result=True,
        log={
            "actor": {"email": "bobert@gmail.com"},
            "events": [
                {
                    "parameters": [
                        {"name": "visibility_change", "value": "external"},
                        {"name": "doc_title", "value": "my shared document"},
                        {"name": "target_domain", "value": "all"},
                        {"name": "visibility", "value": "people_with_link"},
                        {"name": "new_value", "multiValue": ["people_with_link"]},
                    ],
                    "name": "change_document_visibility",
                    "type": "acl_change",
                },
                {
                    "parameters": [{"name": "new_value", "multiValue": ["can_view"]}],
                    "name": "change_document_access_scope",
                    "type": "acl_change",
                },
            ],
            "id": {"applicationName": "drive"},
            "p_row_id": "111222",
        },
    ),
    RuleTest(
        name="Doc Became Public - Link (Allowlisted Domain Not Configured)",
        expected_result=True,
        log={
            "actor": {"email": "bobert@example.com"},
            "events": [
                {
                    "parameters": [
                        {"name": "visibility_change", "value": "external"},
                        {"name": "doc_title", "value": "my shared document"},
                        {"name": "target_domain", "value": "example.com"},
                        {"name": "visibility", "value": "people_within_domain_with_link"},
                        {"name": "new_value", "multiValue": ["people_with_link"]},
                    ],
                    "name": "change_document_visibility",
                    "type": "acl_change",
                },
                {
                    "parameters": [{"name": "new_value", "multiValue": ["can_view"]}],
                    "name": "change_document_access_scope",
                    "type": "acl_change",
                },
            ],
            "id": {"applicationName": "drive"},
            "p_row_id": "111222",
        },
    ),
    RuleTest(
        name="Doc Became Public - Link (Allowlisted Domain Is Configured)",
        expected_result=False,
        mocks=[RuleMock(object_name="ALLOWED_DOMAINS", return_value='[\n  "example.com"\n]')],
        log={
            "actor": {"email": "bobert@example.com"},
            "events": [
                {
                    "parameters": [
                        {"name": "visibility_change", "value": "external"},
                        {"name": "doc_title", "value": "my shared document"},
                        {"name": "target_domain", "value": "example.com"},
                        {"name": "visibility", "value": "people_within_domain_with_link"},
                        {"name": "new_value", "multiValue": ["people_with_link"]},
                    ],
                    "name": "change_document_visibility",
                    "type": "acl_change",
                },
                {
                    "parameters": [{"name": "new_value", "multiValue": ["can_view"]}],
                    "name": "change_document_access_scope",
                    "type": "acl_change",
                },
            ],
            "id": {"applicationName": "drive"},
            "p_row_id": "111222",
        },
    ),
    RuleTest(
        name="Doc Became Private - Link",
        expected_result=False,
        log={
            "actor": {"email": "bobert@example.com"},
            "events": [
                {
                    "parameters": [
                        {"name": "visibility_change", "value": "external"},
                        {"name": "doc_title", "value": "my shared document"},
                        {"name": "target_domain", "value": "all"},
                        {"name": "visibility", "value": "people_with_link"},
                        {"name": "new_value", "multiValue": ["private"]},
                    ],
                    "name": "change_document_visibility",
                    "type": "acl_change",
                },
            ],
            "id": {"applicationName": "drive"},
            "p_row_id": "111222",
        },
    ),
    RuleTest(
        name="Doc Became Public - User",
        expected_result=True,
        log={
            "id": {"applicationName": "drive"},
            "actor": {"email": "bobert@example.com"},
            "kind": "admin#reports#activity",
            "ipAddress": "1.1.1.1",
            "events": [
                {
                    "type": "access",
                    "name": "edit",
                    "parameters": [
                        {"name": "primary_event"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someone@random.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
            ],
        },
    ),
    RuleTest(
        name="Doc Became Public - User (Multiple)",
        expected_result=True,
        log={
            "id": {"applicationName": "drive"},
            "actor": {"email": "bobert@example.com"},
            "kind": "admin#reports#activity",
            "ipAddress": "1.1.1.1",
            "events": [
                {
                    "type": "access",
                    "name": "edit",
                    "parameters": [
                        {"name": "primary_event"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someone@random.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someoneelse@random.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "notbobert@example.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
            ],
        },
    ),
    RuleTest(
        name="Doc Inherits Folder Permissions",
        expected_result=False,
        log={
            "p_row_id": "111222",
            "actor": {"email": "bobert@example.com"},
            "id": {"applicationName": "drive"},
            "events": [
                {
                    "name": "change_user_access_hierarchy_reconciled",
                    "type": "acl_change",
                    "parameters": [{"name": "visibility_change", "value": "internal"}],
                },
            ],
        },
    ),
    RuleTest(
        name="Doc Inherits Folder Permissions - Sharing Link",
        expected_result=False,
        log={
            "p_row_id": "111222",
            "actor": {"email": "bobert@example.com"},
            "id": {"applicationName": "drive"},
            "events": [
                {
                    "name": "change_document_access_scope_hierarchy_reconciled",
                    "type": "acl_change",
                    "parameters": [{"name": "visibility_change", "value": "internal"}],
                },
            ],
        },
    ),
    RuleTest(
        name="Doc Became Public - Public email provider",
        expected_result=True,
        log={
            "id": {"applicationName": "drive"},
            "actor": {"email": "bobert@example.com"},
            "kind": "admin#reports#activity",
            "ipAddress": "1.1.1.1",
            "events": [
                {
                    "type": "access",
                    "name": "edit",
                    "parameters": [
                        {"name": "primary_event"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someone@yandex.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
            ],
        },
    ),
    RuleTest(
        name="Doc Shared With Multiple Users All From ALLOWED_DOMAINS",
        expected_result=False,
        mocks=[RuleMock(object_name="ALLOWED_DOMAINS", return_value='[\n  "example.com", "notexample.com"\n]')],
        log={
            "id": {"applicationName": "drive"},
            "actor": {"email": "bobert@example.com"},
            "kind": "admin#reports#activity",
            "ipAddress": "1.1.1.1",
            "events": [
                {
                    "type": "access",
                    "name": "edit",
                    "parameters": [
                        {"name": "primary_event"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someone@notexample.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "someoneelse@example.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
                {
                    "type": "acl_change",
                    "name": "change_user_access",
                    "parameters": [
                        {"name": "primary_event", "boolValue": True},
                        {"name": "visibility_change", "value": "external"},
                        {"name": "target_user", "value": "notbobert@example.com"},
                        {"name": "old_value", "multiValue": ["none"]},
                        {"name": "new_value", "multiValue": ["can_view"]},
                        {"name": "old_visibility", "value": "people_within_domain_with_link"},
                        {"name": "doc_title", "value": "Hosted Accounts"},
                        {"name": "visibility", "value": "shared_externally"},
                    ],
                },
            ],
        },
    ),
]


@panther_managed
class GSuiteDriveVisibilityChanged(Rule):
    id = "GSuite.DriveVisibilityChanged-prototype"
    display_name = "GSuite External Drive Document"
    enabled = False
    log_types = [LogType.GSUITE_REPORTS]
    tags = ["GSuite", "Collection:Data from Information Repositories", "Configuration Required"]
    reports = {"MITRE ATT&CK": ["TA0009:T1213"]}
    default_severity = Severity.LOW
    default_description = "A Google drive resource became externally accessible.\n"
    default_reference = "https://support.google.com/a/users/answer/12380484?hl=en&sjid=864417124752637253-EU"
    default_runbook = "Investigate whether the drive document is appropriate to be publicly accessible.\n"
    summary_attributes = ["actor:email"]
    dedup_period_minutes = 360
    tests = g_suite_drive_visibility_changed_tests
    # Add any domain name(s) that you expect to share documents with in the ALLOWED_DOMAINS set
    ALLOWED_DOMAINS = set()
    PUBLIC_PROVIDERS = {
        "gmail.com",
        "yahoo.com",
        "outlook.com",
        "aol.com",
        "yandex.com",
        "protonmail.com",
        "pm.me",
        "icloud.com",
        "tutamail.com",
        "tuta.io",
        "keemail.me",
        "mail.com",
        "zohomail.com",
        "hotmail.com",
        "msn.com",
    }
    VISIBILITY = {
        "people_with_link",
        "people_within_domain_with_link",
        "public_on_the_web",
        "shared_externally",
        "unknown",
    }
    ALERT_DETAILS = {}
    # Events where documents have changed perms due to parent folder change
    INHERITANCE_EVENTS = {
        "change_user_access_hierarchy_reconciled",
        "change_document_access_scope_hierarchy_reconciled",
    }

    def init_alert_details(self, log):
        self.ALERT_DETAILS[log] = {
            "ACCESS_SCOPE": "<UNKNOWN_ACCESS_SCOPE>",
            "DOC_TITLE": "<UNKNOWN_TITLE>",
            "NEW_VISIBILITY": "<UNKNOWN_VISIBILITY>",
            "TARGET_USER_EMAILS": ["<UNKNOWN_USER>"],
            "TARGET_DOMAIN": "<UNKNOWN_DOMAIN>",
        }

    def user_is_external(self, target_user):
        # We need to type-cast ALLOWED_DOMAINS for unit testing mocks
        if isinstance(self.ALLOWED_DOMAINS, MagicMock):
            self.ALLOWED_DOMAINS = set(json.loads(self.ALLOWED_DOMAINS()))  # pylint: disable=not-callable
        for domain in self.ALLOWED_DOMAINS:
            if domain in target_user:
                return False
        return True

    def rule(self, event):
        # pylint: disable=too-complex
        if deep_get(event, "id", "applicationName") != "drive":
            return False
        # Events that have the types in INHERITANCE_EVENTS are
        # changes to documents and folders that occur due to
        # a change in the parent folder's permission. We ignore
        # these events to prevent every folder change from
        # generating multiple alerts.
        if deep_get(event, "events", "name") in self.INHERITANCE_EVENTS:
            return False
        log = event.get("p_row_id")
        self.init_alert_details(log)
        #########
        # for visibility changes that apply to a domain, not a user
        change_document_visibility = False
        # We need to type-cast ALLOWED_DOMAINS for unit testing mocks
        if isinstance(self.ALLOWED_DOMAINS, MagicMock):
            self.ALLOWED_DOMAINS = set(json.loads(self.ALLOWED_DOMAINS()))  # pylint: disable=not-callable
        for details in event.get("events", [{}]):
            if (
                details.get("type") == "acl_change"
                and details.get("name") == "change_document_visibility"
                and (param_lookup(details.get("parameters", {}), "new_value") != ["private"])
                and (param_lookup(details.get("parameters", {}), "target_domain") not in self.ALLOWED_DOMAINS)
                and (param_lookup(details.get("parameters", {}), "visibility") in self.VISIBILITY)
            ):
                self.ALERT_DETAILS[log]["TARGET_DOMAIN"] = param_lookup(details.get("parameters", {}), "target_domain")
                self.ALERT_DETAILS[log]["NEW_VISIBILITY"] = param_lookup(details.get("parameters", {}), "visibility")
                self.ALERT_DETAILS[log]["DOC_TITLE"] = param_lookup(details.get("parameters", {}), "doc_title")
                change_document_visibility = True
                break
        # "change_document_visibility" events are always paired with
        # "change_document_access_scope" events. the "target_domain" and
        # "visibility" attributes are equivalent.
        if change_document_visibility:
            for details in event.get("events", [{}]):
                if (
                    details.get("type") == "acl_change"
                    and details.get("name") == "change_document_access_scope"
                    and (param_lookup(details.get("parameters", {}), "new_value") != ["none"])
                ):
                    self.ALERT_DETAILS[log]["ACCESS_SCOPE"] = param_lookup(details.get("parameters", {}), "new_value")
            return True
        #########
        # for visibility changes that apply to a user
        # there is a change_user_access event for each user
        # change_user_access and change_document_visibility events are
        # not found in the same report
        change_user_access = False
        for details in event.get("events", [{}]):
            if (
                details.get("type") == "acl_change"
                and details.get("name") == "change_user_access"
                and (param_lookup(details.get("parameters", {}), "new_value") != ["none"])
                and self.user_is_external(param_lookup(details.get("parameters", {}), "target_user"))
            ):
                if self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"] != ["<UNKNOWN_USER>"]:
                    self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"].append(
                        param_lookup(details.get("parameters", {}), "target_user"),
                    )
                else:
                    self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"] = [
                        param_lookup(details.get("parameters", {}), "target_user"),
                    ]
                    self.ALERT_DETAILS[log]["DOC_TITLE"] = param_lookup(details.get("parameters", {}), "doc_title")
                    self.ALERT_DETAILS[log]["ACCESS_SCOPE"] = param_lookup(details.get("parameters", {}), "new_value")
                change_user_access = True
        return change_user_access

    def alert_context(self, event):
        log = event.get("p_row_id")
        if self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"] != ["<UNKNOWN_USER>"]:
            return {"target users": self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"]}
        return {}

    def dedup(self, event):
        return deep_get(event, "actor", "email", default="<UNKNOWN_USER>")

    def title(self, event):
        log = event.get("p_row_id")
        if self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"] != ["<UNKNOWN_USER>"]:
            if len(self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"]) == 1:
                sharing_scope = self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"][0]
            else:
                sharing_scope = "multiple users"
            if self.ALERT_DETAILS[log]["NEW_VISIBILITY"] == "shared_externally":
                sharing_scope += " (outside the document's current domain)"
        elif self.ALERT_DETAILS[log]["TARGET_DOMAIN"] == "all":
            sharing_scope = "the entire internet"
            if self.ALERT_DETAILS[log]["NEW_VISIBILITY"] == "people_with_link":
                sharing_scope += " (anyone with the link)"
            elif self.ALERT_DETAILS[log]["NEW_VISIBILITY"] == "public_on_the_web":
                sharing_scope += " (link not required)"
        else:
            sharing_scope = f"the {self.ALERT_DETAILS[log]['TARGET_DOMAIN']} domain"
            if self.ALERT_DETAILS[log]["NEW_VISIBILITY"] == "people_within_domain_with_link":
                sharing_scope += f" (anyone in {self.ALERT_DETAILS[log]['TARGET_DOMAIN']} with the link)"
            elif self.ALERT_DETAILS[log]["NEW_VISIBILITY"] == "public_in_the_domain":
                sharing_scope += f" (anyone in {self.ALERT_DETAILS[log]['TARGET_DOMAIN']})"
        # alert_access_scope = ALERT_DETAILS[log]["ACCESS_SCOPE"][0].replace("can_", "")
        return f"User [{deep_get(event, 'actor', 'email', default='<UNKNOWN_USER>')}] made documents externally visible"

    def severity(self, event):
        log = event.get("p_row_id")
        if self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"] != ["<UNKNOWN_USER>"]:
            for address in self.ALERT_DETAILS[log]["TARGET_USER_EMAILS"]:
                domain = address.split("@")[1]
                if domain in self.PUBLIC_PROVIDERS:
                    return "LOW"
        return "INFO"



"""get_info, get_k8s_info, get_flow_logs_info etc return dicts of the most commonly used fields."""

from pypanther.helpers.panther_base_helpers import deep_get


def get_info(event):
    fields = {
        "principal": "protoPayload.authenticationInfo.principalEmail",
        "project_id": "protoPayload.resource.labels.project_id",
        "caller_ip": "protoPayload.requestMetadata.callerIP",
        "user_agent": "protoPayload.requestMetadata.callerSuppliedUserAgent",
        "method_name": "protoPayload.methodName",
    }
    return {name: deep_get(event, *(path.split("."))) for name, path in fields.items()}


def get_k8s_info(event):
    """
    Get GCP K8s info such as pod, authorized user etc.
    return a tuple of strings
    """
    pod_slug = deep_get(event, "protoPayload", "resourceName")
    # core/v1/namespaces/<namespace>/pods/<pod-id>/<action>
    _, _, _, namespace, _, pod, _ = pod_slug.split("/")
    return get_info(event) | {"namespace": namespace, "pod": pod}


def get_flow_log_info(event):
    fields = {
        "src_ip": "jsonPayload.connection.src_ip",
        "dest_ip": "jsonPayload.connection.dest_ip",
        "src_port": "jsonPayload.connection.src_port",
        "dest_port": "jsonPayload.connection.dest_port",
        "protocol": "jsonPayload.connection.protocol",
        "bytes_sent": "jsonPayload.bytes_sent",
        "reporter": "jsonPayload.reporter",
    }
    return {name: deep_get(event, *(path.split("."))) for name, path in fields.items()}


def gcp_alert_context(event):
    return {
        "project": deep_get(event, "resource", "labels", "project_id", default=""),
        "principal": deep_get(event, "protoPayload", "authenticationInfo", "principalEmail", default=""),
        "caller_ip": deep_get(event, "protoPayload", "requestMetadata", "callerIP", default=""),
        "methodName": deep_get(event, "protoPayload", "methodName", default=""),
        "resourceName": deep_get(event, "protoPayload", "resourceName", default=""),
        "serviceName": deep_get(event, "protoPayload", "serviceName", default=""),
    }



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import aws_rule_context

awsip_set_modified_tests: list[RuleTest] = [
    RuleTest(
        name="CreateIPSet Event",
        expected_result=True,
        log={
            "awsregion": "us-east-1",
            "eventid": "abc-123",
            "eventname": "CreateIPSet",
            "eventsource": "guardduty.amazonaws.com",
            "eventtime": "2022-07-17 04:50:23",
            "eventtype": "AwsApiCall",
            "eventversion": "1.08",
            "p_any_aws_instance_ids": ["testinstanceid"],
            "p_event_time": "2022-07-17 04:50:23",
            "p_log_type": "AWS.CloudTrail",
            "p_parse_time": "2022-07-17 04:55:11.788",
            "recipientAccountId": "123456789012",
        },
    ),
    RuleTest(
        name="UpdateIPSet",
        expected_result=True,
        log={
            "awsregion": "us-east-1",
            "eventid": "abc-123",
            "eventname": "CreateIPSet",
            "eventsource": "guardduty.amazonaws.com",
            "eventtime": "2022-07-17 04:50:23",
            "eventtype": "AwsApiCall",
            "eventversion": "1.08",
            "p_any_aws_instance_ids": ["testinstanceid"],
            "p_event_time": "2022-07-17 04:50:23",
            "p_log_type": "AWS.CloudTrail",
            "p_parse_time": "2022-07-17 04:55:11.788",
            "recipientAccountId": "123456789012",
        },
    ),
    RuleTest(
        name="NotIPSet",
        expected_result=False,
        log={
            "awsregion": "us-east-1",
            "eventid": "abc-123",
            "eventname": "ModifyInstanceAttributes",
            "eventsource": "guardduty.amazonaws.com",
            "eventtime": "2022-07-17 04:50:23",
            "eventtype": "AwsApiCall",
            "eventversion": "1.08",
            "p_any_aws_instance_ids": ["testinstanceid"],
            "p_event_time": "2022-07-17 04:50:23",
            "p_log_type": "AWS.CloudTrail",
            "p_parse_time": "2022-07-17 04:55:11.788",
            "recipientAccountId": "123456789012",
        },
    ),
]


@panther_managed
class AWSIPSetModified(Rule):
    default_description = "Detects creation and updates of the list of trusted IPs used by GuardDuty and WAF. Potentially to disable security alerts against malicious IPs."
    display_name = "AWS Trusted IPSet Modified"
    reports = {"MITRE ATT&CK": ["TA0005:T1562"]}
    default_reference = "https://docs.aws.amazon.com/managedservices/latest/ctref/management-monitoring-guardduty-ip-set-update-review-required.html"
    default_severity = Severity.HIGH
    log_types = [LogType.AWS_CLOUDTRAIL]
    id = "AWS.IPSet.Modified-prototype"
    tests = awsip_set_modified_tests
    IPSET_ACTIONS = ["CreateIPSet", "UpdateIPSet"]

    def rule(self, event):
        if (
            event.get("eventSource", "") == "guardduty.amazonaws.com"
            or event.get("eventSource", "") == "wafv2.amazonaws.com"
        ):
            if event.get("eventName", "") in self.IPSET_ACTIONS:
                return True
        return False

    def title(self, event):
        return f"IPSet was modified in [{event.get('recipientAccountId', '')}]"

    def alert_context(self, event):
        return aws_rule_context(event)



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.gcp_base_helpers import gcp_alert_context
from pypanther.helpers.panther_base_helpers import deep_get, deep_walk

gcp_cloud_build_potential_privilege_escalation_tests: list[RuleTest] = [
    RuleTest(
        name="GCP CloudBuild - Build with Potentially Privileged Access",
        expected_result=True,
        log={
            "logName": "projects/some-project/logs/cloudaudit.googleapis.com%2Factivity",
            "operation": {
                "first": True,
                "id": "operations/build/some-project/YzNhZWI0YWYtNjAwNi00YzM5LTgxYmUtMjhmMjc1YzJkOGEz",
                "producer": "cloudbuild.googleapis.com",
            },
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {
                    "principalEmail": "whodoneit@some-project.iam.gserviceaccount.com",
                    "principalSubject": "serviceAccount:whodoneit@some-project.iam.gserviceaccount.com",
                    "serviceAccountKeyName": "//iam.googleapis.com/projects/some-project/serviceAccounts/whodoneit@some-project.iam.gserviceaccount.com/keys/123er456788",
                },
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "cloudbuild.builds.create",
                        "resource": "projects/some-project",
                        "resourceAttributes": {},
                    },
                ],
                "methodName": "google.devtools.cloudbuild.v1.CloudBuild.CreateBuild",
                "request": {
                    "@type": "type.googleapis.com/google.devtools.cloudbuild.v1.CreateBuildRequest",
                    "build": {},
                    "projectId": "some-project",
                },
                "requestMetadata": {
                    "callerIP": "189.163.74.177",
                    "callerSuppliedUserAgent": "(gzip),gzip(gfe),gzip(gfe)",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2024-01-25T11:55:09.740095Z"},
                },
                "resourceLocation": {"currentLocations": ["global"]},
                "resourceName": "projects/some-project/builds",
                "serviceName": "cloudbuild.googleapis.com",
            },
            "receiveTimestamp": "2024-01-25 11:55:09.854909113",
            "resource": {
                "labels": {
                    "build_id": "c3aeb4ap-6006-4c39-81be-28f275c2d8a3",
                    "build_trigger_id": "",
                    "project_id": "some-project",
                },
                "type": "build",
            },
            "severity": "NOTICE",
            "timestamp": "2024-01-25 11:55:08.919358000",
        },
    ),
    RuleTest(
        name="GCP CreateBrand - No Privileged Access",
        expected_result=False,
        log={
            "logName": "projects/some-project/logs/cloudaudit.googleapis.com%2Factivity",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {"principalEmail": "john.doe@some-project.com"},
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "clientauthconfig.brands.create",
                        "resource": "brands/1028347248702",
                        "resourceAttributes": {},
                    },
                ],
                "methodName": "CreateBrand",
                "request": {
                    "brand": {
                        "displayName": "NewBrand",
                        "projectNumbers": ["1028345275902"],
                        "supportEmail": "john.doe@some-project.com",
                    },
                    "visibility": "INTERNAL",
                },
                "requestMetadata": {
                    "callerIP": "189.163.74.177",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2024-01-24T14:30:05.891336Z"},
                },
                "resourceName": "brands/1028347275902",
                "response": {
                    "brandId": "1028347248702",
                    "creationTime": "2024-01-24T14:30:05.400Z",
                    "displayName": "NewBrand test",
                    "projectNumbers": ["1028347248702"],
                    "supportEmail": "some.user@company.com",
                    "updateTime": "2024-01-24T14:30:05.866002Z",
                },
                "serviceName": "clientauthconfig.googleapis.com",
                "status": {},
            },
            "receiveTimestamp": "2024-01-24 14:30:06.741210353",
            "resource": {
                "labels": {"brand_id": "1028347248702", "project_id": "some-project"},
                "type": "client_auth_config_brand",
            },
            "severity": "NOTICE",
            "timestamp": "2024-01-24 14:30:05.207884000",
        },
    ),
]


@panther_managed
class GCPCloudBuildPotentialPrivilegeEscalation(Rule):
    log_types = [LogType.GCP_AUDIT_LOG]
    default_description = "Detects privilege escalation attacks designed to gain access to the Cloud Build Service Account. A user with permissions to start a new build with Cloud Build can gain access to the Cloud Build Service Account and abuse it for more access to the environment."
    display_name = "GCP CloudBuild Potential Privilege Escalation"
    id = "GCP.CloudBuild.Potential.Privilege.Escalation-prototype"
    default_reference = "https://rhinosecuritylabs.com/gcp/iam-privilege-escalation-gcp-cloudbuild/"
    default_runbook = "Confirm this was authorized and necessary behavior. To defend against this privilege escalation attack, it is necessary to restrict the permissions granted to the Cloud Build Service Account and to be careful granting the cloudbuild.builds.create permission to any users in your Organization. Most importantly, you need to know that any user who is granted cloudbuild.builds.create, is also indirectly granted all the permissions granted to the Cloud Build Service Account. If that’s alright with you, then you may not need to worry about this attack vector, but it is still highly recommended to modify the default permissions granted to the Cloud Build Service Account."
    reports = {"MITRE ATT&CK": ["TA0004:T1548"]}
    default_severity = Severity.HIGH
    tests = gcp_cloud_build_potential_privilege_escalation_tests

    def rule(self, event):
        if not deep_get(event, "protoPayload", "methodName", default="METHOD_NOT_FOUND").endswith(
            "CloudBuild.CreateBuild",
        ):
            return False
        authorization_info = deep_walk(event, "protoPayload", "authorizationInfo")
        if not authorization_info:
            return False
        for auth in authorization_info:
            if auth.get("permission") == "cloudbuild.builds.create" and auth.get("granted") is True:
                return True
        return False

    def title(self, event):
        actor = deep_get(event, "protoPayload", "authenticationInfo", "principalEmail", default="<ACTOR_NOT_FOUND>")
        operation = deep_get(event, "protoPayload", "methodName", default="<OPERATION_NOT_FOUND>")
        project_id = deep_get(event, "resource", "labels", "project_id", default="<PROJECT_NOT_FOUND>")
        return f"[GCP]: [{actor}] performed [{operation}] on project [{project_id}]"

    def alert_context(self, event):
        return gcp_alert_context(event)



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get

g_suite_group_banned_user_tests: list[RuleTest] = [
    RuleTest(
        name="User Added",
        expected_result=False,
        log={
            "id": {"applicationName": "groups_enterprise"},
            "actor": {"email": "homer.simpson@example.com"},
            "type": "moderator_action",
            "name": "add_member",
        },
    ),
    RuleTest(
        name="User Banned from Group",
        expected_result=True,
        log={
            "id": {"applicationName": "groups_enterprise"},
            "actor": {"email": "homer.simpson@example.com"},
            "type": "moderator_action",
            "name": "ban_user_with_moderation",
        },
    ),
]


@panther_managed
class GSuiteGroupBannedUser(Rule):
    id = "GSuite.GroupBannedUser-prototype"
    display_name = "GSuite User Banned from Group"
    log_types = [LogType.GSUITE_ACTIVITY_EVENT]
    tags = ["GSuite"]
    default_severity = Severity.LOW
    default_description = "A GSuite user was banned from an enterprise group by moderator action.\n"
    default_reference = "https://support.google.com/a/users/answer/9303224?hl=en&sjid=864417124752637253-EU"
    default_runbook = "Investigate the banned user to see if further disciplinary action needs to be taken.\n"
    summary_attributes = ["actor:email"]
    tests = g_suite_group_banned_user_tests

    def rule(self, event):
        if deep_get(event, "id", "applicationName") != "groups_enterprise":
            return False
        if event.get("type") == "moderator_action":
            return bool(event.get("name") == "ban_user_with_moderation")
        return False

    def title(self, event):
        return (
            f"User [{deep_get(event, 'actor', 'email', default='<UNKNOWN_EMAIL>')}] banned another user from a group."
        )



from pypanther.base import DataModel, DataModelMapping, LogType


class StandardSlackIntegrationLogs(DataModel):
    id: str = "Standard.Slack.IntegrationLogs"
    display_name: str = "Slack Integration Logs"
    enabled: bool = True
    log_types: list[str] = [LogType.SLACK_INTEGRATION_LOGS]
    mappings: list[DataModelMapping] = [DataModelMapping(name="actor_user", path="user_name")]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get

g_suite_workspace_calendar_external_sharing_setting_tests: list[RuleTest] = [
    RuleTest(
        name="Admin Set Default Calendar SHARING_OUTSIDE_DOMAIN Setting to READ_ONLY_ACCESS",
        expected_result=True,
        log={
            "actor": {"callerType": "USER", "email": "example@example.io", "profileId": "12345"},
            "id": {
                "applicationName": "admin",
                "customerId": "D12345",
                "time": "2022-12-11 01:06:26.303000000",
                "uniqueQualifier": "-12345",
            },
            "ipAddress": "12.12.12.12",
            "kind": "admin#reports#activity",
            "name": "CHANGE_CALENDAR_SETTING",
            "parameters": {
                "DOMAIN_NAME": "example.io",
                "NEW_VALUE": "READ_ONLY_ACCESS",
                "OLD_VALUE": "DEFAULT",
                "ORG_UNIT_NAME": "Example IO",
                "SETTING_NAME": "SHARING_OUTSIDE_DOMAIN",
            },
            "type": "CALENDAR_SETTINGS",
        },
    ),
    RuleTest(
        name="Admin Set Default Calendar SHARING_OUTSIDE_DOMAIN Setting to READ_WRITE_ACCESS",
        expected_result=True,
        log={
            "actor": {"callerType": "USER", "email": "example@example.io", "profileId": "12345"},
            "id": {
                "applicationName": "admin",
                "customerId": "D12345",
                "time": "2022-12-11 01:06:26.303000000",
                "uniqueQualifier": "-12345",
            },
            "ipAddress": "12.12.12.12",
            "kind": "admin#reports#activity",
            "name": "CHANGE_CALENDAR_SETTING",
            "parameters": {
                "DOMAIN_NAME": "example.io",
                "NEW_VALUE": "READ_WRITE_ACCESS",
                "OLD_VALUE": "READ_ONLY_ACCESS",
                "ORG_UNIT_NAME": "Example IO",
                "SETTING_NAME": "SHARING_OUTSIDE_DOMAIN",
            },
            "type": "CALENDAR_SETTINGS",
        },
    ),
    RuleTest(
        name="Admin Set Default Calendar SHARING_OUTSIDE_DOMAIN Setting to MANAGE_ACCESS",
        expected_result=True,
        log={
            "actor": {"callerType": "USER", "email": "example@example.io", "profileId": "12345"},
            "id": {
                "applicationName": "admin",
                "customerId": "D12345",
                "time": "2022-12-11 01:06:26.303000000",
                "uniqueQualifier": "-12345",
            },
            "ipAddress": "12.12.12.12",
            "kind": "admin#reports#activity",
            "name": "CHANGE_CALENDAR_SETTING",
            "parameters": {
                "DOMAIN_NAME": "example.io",
                "NEW_VALUE": "MANAGE_ACCESS",
                "OLD_VALUE": "READ_WRITE_ACCESS",
                "ORG_UNIT_NAME": "Example IO",
                "SETTING_NAME": "SHARING_OUTSIDE_DOMAIN",
            },
            "type": "CALENDAR_SETTINGS",
        },
    ),
    RuleTest(
        name="Non-Default Calendar SHARING_OUTSIDE_DOMAIN event",
        expected_result=False,
        log={
            "actor": {"callerType": "USER", "email": "user@example.io", "profileId": "111111111111111111111"},
            "id": {
                "applicationName": "admin",
                "customerId": "D12345",
                "time": "2022-12-12 22:21:40.106000000",
                "uniqueQualifier": "1000000000000000000",
            },
            "kind": "admin#reports#activity",
            "name": "CUSTOMER_TAKEOUT_SUCCEEDED",
            "parameters": {"OBFUSCATED_CUSTOMER_TAKEOUT_REQUEST_ID": "00mmmmmmmmmmmmm"},
            "type": "CUSTOMER_TAKEOUT",
        },
    ),
    RuleTest(
        name="ListObject Type",
        expected_result=False,
        log={
            "actor": {"email": "user@example.io", "profileId": "118111111111111111111"},
            "id": {
                "applicationName": "drive",
                "customerId": "D12345",
                "time": "2022-12-20 17:27:47.080000000",
                "uniqueQualifier": "-7312729053723258069",
            },
            "ipAddress": "12.12.12.12",
            "kind": "admin#reports#activity",
            "name": "rename",
            "parameters": {
                "actor_is_collaborator_account": None,
                "billable": True,
                "doc_id": "1GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG",
                "doc_title": "Document Title- Found Here",
                "doc_type": "presentation",
                "is_encrypted": None,
                "new_value": ["Document Title- Found Here"],
                "old_value": ["Document Title- Old"],
                "owner": "user@example.io",
                "owner_is_shared_drive": None,
                "owner_is_team_drive": None,
                "primary_event": True,
                "visibility": "private",
            },
            "type": "access",
        },
    ),
]


@panther_managed
class GSuiteWorkspaceCalendarExternalSharingSetting(Rule):
    id = "GSuite.Workspace.CalendarExternalSharingSetting-prototype"
    display_name = "GSuite Workspace Calendar External Sharing Setting Change"
    log_types = [LogType.GSUITE_ACTIVITY_EVENT]
    tags = ["GSuite"]
    reports = {"MITRE ATT&CK": ["TA0007:T1087"]}
    default_severity = Severity.MEDIUM
    default_description = "A Workspace Admin Changed The Sharing Settings for Primary Calendars\n"
    default_reference = "https://support.google.com/a/answer/60765?hl=en"
    default_runbook = "Restore the calendar sharing setting to the previous value. If unplanned, use indicator search to identify other activity from this administrator.\n"
    summary_attributes = ["actor:email"]
    tests = g_suite_workspace_calendar_external_sharing_setting_tests

    def rule(self, event):
        if not all(
            [
                event.get("name", "") == "CHANGE_CALENDAR_SETTING",
                deep_get(event, "parameters", "SETTING_NAME", default="") == "SHARING_OUTSIDE_DOMAIN",
            ],
        ):
            return False
        return deep_get(event, "parameters", "NEW_VALUE", default="") in [
            "READ_WRITE_ACCESS",
            "READ_ONLY_ACCESS",
            "MANAGE_ACCESS",
        ]

    def title(self, event):
        return f"GSuite workspace setting for default calendar sharing was changed by [{deep_get(event, 'actor', 'email', default='<UNKNOWN_EMAIL>')}] from [{deep_get(event, 'parameters', 'OLD_VALUE', default='<NO_OLD_SETTING_FOUND>')}] to [{deep_get(event, 'parameters', 'NEW_VALUE', default='<NO_NEW_SETTING_FOUND>')}]"



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed

teleport_network_scanning_tests: list[RuleTest] = [
    RuleTest(
        name="Echo command",
        expected_result=False,
        log={
            "argv": [],
            "cgroup_id": 4294967537,
            "code": "T4000I",
            "ei": 15,
            "event": "session.command",
            "login": "root",
            "namespace": "default",
            "path": "/bin/echo",
            "pid": 7143,
            "ppid": 7115,
            "program": "echo",
            "return_code": 0,
            "server_id": "e75992b4-9e27-456f-b1c9-7a32da83c661",
            "sid": "8a3fc038-785b-43f3-8737-827b3e25fe5b",
            "time": "2020-08-17T17:40:37.491Z",
            "uid": "8eaf8f39-09d4-4a42-a22a-65163d2af702",
            "user": "panther",
        },
    ),
    RuleTest(
        name="Nmap with no args",
        expected_result=False,
        log={
            "argv": [],
            "cgroup_id": 4294967672,
            "code": "T4000I",
            "ei": 16,
            "event": "session.command",
            "login": "root",
            "namespace": "default",
            "path": "/bin/nmap",
            "pid": 13555,
            "ppid": 13525,
            "program": "nmap",
            "return_code": 0,
            "server_id": "e75992b4-9e27-456f-b1c9-7a32da83c661",
            "sid": "a3562a0e-e57f-4273-9f69-eedb6cd029cb",
            "time": "2020-08-17T21:13:47.117Z",
            "uid": "c7f6367b-04bb-4b1d-9a3a-0497e8f4a650",
            "user": "panther",
        },
    ),
    RuleTest(
        name="Nmap with args",
        expected_result=True,
        log={
            "argv": ["-v", "-iR", "100000", "-Pn", "-p", "80"],
            "cgroup_id": 4294967672,
            "code": "T4000I",
            "ei": 16,
            "event": "session.command",
            "login": "root",
            "namespace": "default",
            "path": "/bin/nmap",
            "pid": 13555,
            "ppid": 13525,
            "program": "nmap",
            "return_code": 0,
            "server_id": "e75992b4-9e27-456f-b1c9-7a32da83c661",
            "sid": "a3562a0e-e57f-4273-9f69-eedb6cd029cb",
            "time": "2020-08-17T21:13:47.117Z",
            "uid": "c7f6367b-04bb-4b1d-9a3a-0497e8f4a650",
            "user": "panther",
        },
    ),
    RuleTest(
        name="Nmap running from crontab",
        expected_result=True,
        log={
            "cgroup_id": 4294967792,
            "code": "T4002I",
            "dst_addr": "67.205.137.100",
            "dst_port": 1723,
            "ei": 32,
            "event": "session.network",
            "login": "root",
            "namespace": "default",
            "pid": 15412,
            "program": "nmap",
            "server_id": "e75992b4-9e27-456f-b1c9-7a32da83c661",
            "sid": "a3562a0e-e57f-4273-9f69-eedb6cd029cb",
            "src_addr": "172.31.9.159",
            "time": "2020-08-18T17:37:35.883Z",
            "uid": "3e067d21-a5fb-47a3-af09-e6b9da39753c",
            "user": "panther",
            "version": 4,
        },
    ),
]


@panther_managed
class TeleportNetworkScanning(Rule):
    id = "Teleport.NetworkScanning-prototype"
    display_name = "Teleport Network Scan Initiated"
    log_types = [LogType.GRAVITATIONAL_TELEPORT_AUDIT]
    tags = ["SSH", "Discovery:Network Service Discovery"]
    default_severity = Severity.MEDIUM
    default_description = (
        "A user has invoked a network scan that could potentially indicate enumeration of the network."
    )
    reports = {"MITRE ATT&CK": ["TA0007:T1046"]}
    default_reference = "https://goteleport.com/docs/management/admin/"
    default_runbook = "Find related commands within the time window and determine if the command was invoked legitimately. Examine the arguments to determine how the command was used.\n"
    summary_attributes = ["event", "code", "user", "program", "path", "return_code", "login", "server_id", "sid"]
    tests = teleport_network_scanning_tests
    SCAN_COMMANDS = {"arp", "arp-scan", "fping", "nmap"}

    def rule(self, event):
        # Filter out commands
        if event.get("event") == "session.command" and (not event.get("argv")):
            return False
        # Check that the program is in our watch list
        return event.get("program") in self.SCAN_COMMANDS

    def title(self, event):
        return f"User [{event.get('user', '<UNKNOWN_USER>')}] has issued a network scan with [{event.get('program', '<UNKNOWN_PROGRAM>')}] on [{event.get('cluster_name', '<UNKNOWN_CLUSTER>')}]"



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get
from pypanther.helpers.panther_notion_helpers import notion_alert_context

notion_workspace_public_page_added_tests: list[RuleTest] = [
    RuleTest(
        name="Public page added",
        expected_result=True,
        log={
            "event": {
                "id": "eeeeeeee-dddd-4444-bbbb-4444444444444",
                "timestamp": "2023-05-15T19:14:21.031Z",
                "workspace_id": "vvvvvvvv-dddd-4444-bbbb-666666666666",
                "actor": {
                    "object": "user",
                    "id": "44444444-cccc-7777-aaaa-666666666666",
                    "type": "person",
                    "person": {"email": "example@personemail.com"},
                },
                "ip_address": "00.000.00.000",
                "platform": "web",
                "type": "workspace.settings.public_homepage_added",
                "workspace.settings.public_homepage_added": {
                    "new_public_page": {"type": "database_id", "database_id": "4b801dc7-d724-4fbb-afd0-9885cbc12405"},
                },
            },
        },
    ),
    RuleTest(
        name="Workspace Exported",
        expected_result=False,
        log={
            "event": {
                "actor": {
                    "id": "bd37477c-869d-418b-abdb-0fc727b38b5e",
                    "object": "user",
                    "person": {"email": "homer.simpson@yourcompany.io"},
                    "type": "person",
                },
                "details": {
                    "parent": {"type": "workspace_id", "workspace_id": "ab99as87-6abc-4dcf-808b-111999882299"},
                    "target": {"page_id": "3cd2c560-d1b9-474e-b46e-gh8899002763", "type": "page_id"},
                },
                "id": "d4b9963f-12a8-4b01-b597-233a140abf5e",
                "ip_address": "12.12.12.12",
                "platform": "web",
                "timestamp": "2023-06-01 18:57:07.486000000",
                "type": "page.exported",
                "workspace_id": "ea65b016-6abc-4dcf-808b-e119617b55d1",
            },
        },
    ),
]


@panther_managed
class NotionWorkspacePublicPageAdded(Rule):
    id = "Notion.Workspace.Public.Page.Added-prototype"
    display_name = "Notion Workspace public page added"
    log_types = [LogType.NOTION_AUDIT_LOGS]
    tags = ["Notion", "Data Security", "Information Disclosure"]
    default_severity = Severity.INFO
    default_description = "A Notion page was set to public in your worksace."
    default_runbook = "A Notion page was made public. Check with the author to determine why this page was made public."
    default_reference = "https://www.notion.so/help/public-pages-and-web-publishing"
    tests = notion_workspace_public_page_added_tests

    def rule(self, event):
        event_type = event.deep_get("event", "type", default="<NO_EVENT_TYPE_FOUND>")
        return event_type == "workspace.settings.public_homepage_added"

    def title(self, event):
        actor = event.deep_get("event", "actor", "person", "email", default="<NO_EMAIL_FOUND>")
        wkspc_id = event.deep_get("event", "workspace_id", default="<NO_WORKSPACE_ID_FOUND>")
        db_id = deep_get(
            event,
            "event",
            "workspace.settings.public_homepage_added",
            "new_public_page",
            "database_id",
            default="<NO_DATABASE_ID_FOUND>",
        )
        return f"Notion User [{actor}] added a new public homepage [{db_id}] in workspace [{wkspc_id}]"

    def alert_context(self, event):
        context = notion_alert_context(event)
        workspace_id = event.deep_get("event", "workspace_id", default="<NO_WORKSPACE_ID_FOUND>")
        db_id = deep_get(
            event,
            "event",
            "workspace.settings.public_homepage_added",
            "new_public_page",
            "database_id",
            default="<NO_DATABASE_ID_FOUND>",
        )
        context["workspace_id"] = workspace_id
        context["page_id"] = db_id
        return context



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get

gcp_service_accountor_keys_created_tests: list[RuleTest] = [
    RuleTest(
        name="Created Service Account Key",
        expected_result=True,
        log={
            "insertId": "1iyadj0d5bmj",
            "logName": "projects/gcp-project1/logs/cloudaudit.googleapis.com%2Factivity",
            "p_any_ip_addresses": ["1.2.3.4"],
            "p_event_time": "2023-03-09 15:50:36.148",
            "p_log_type": "GCP.AuditLog",
            "p_parse_time": "2023-03-09 15:52:14.605",
            "p_row_id": "eee460dd2ac5b8a9a6cdfded16ff75",
            "p_source_id": "4fc88a5a-2d51-4279-9c4a-08fa7cc52566",
            "p_source_label": "gcplogsource",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {
                    "principalEmail": "staging@company.io",
                    "principalSubject": "user:staging@company.io",
                },
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "iam.serviceAccountKeys.create",
                        "resource": "projects/-/serviceAccounts/123456789098765434567",
                        "resourceAttributes": {"name": "projects/-/serviceAccounts/123456789098765434567"},
                    },
                ],
                "methodName": "google.iam.admin.v1.CreateServiceAccountKey",
                "request": {
                    "@type": "type.googleapis.com/google.iam.admin.v1.CreateServiceAccountKeyRequest",
                    "name": "projects/gcp-project1/serviceAccounts/123456789098765434567",
                    "private_key_type": 2,
                },
                "requestMetadata": {
                    "callerIP": "1.2.3.4",
                    "callerSuppliedUserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36,gzip(gfe)",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2023-03-09T15:50:36.163986905Z"},
                },
                "resourceName": "projects/-/serviceAccounts/123456789098765434567",
                "response": {
                    "@type": "type.googleapis.com/google.iam.admin.v1.ServiceAccountKey",
                    "key_algorithm": 2,
                    "key_origin": 2,
                    "key_type": 1,
                    "name": "projects/gcp-project1/serviceAccounts/created-service-account@gcp-project1.iam.gserviceaccount.com/keys/69049163b190437ab9279b28e8afa24bb5c5b076",
                    "private_key_type": 2,
                    "valid_after_time": {"seconds": 1678377036.0},
                    "valid_before_time": {"seconds": 253402300799.0},
                },
                "serviceName": "iam.googleapis.com",
                "status": {},
            },
            "receiveTimestamp": "2023-03-09 15:50:37.603",
            "resource": {
                "labels": {
                    "email_id": "created-service-account@gcp-project1.iam.gserviceaccount.com",
                    "project_id": "gcp-project1",
                    "unique_id": "123456789098765434567",
                },
                "type": "service_account",
            },
            "severity": "NOTICE",
            "timestamp": "2023-03-09 15:50:36.148",
        },
    ),
    RuleTest(
        name="Created Service Account",
        expected_result=True,
        log={
            "insertId": "1iyadj0d5bcq",
            "logName": "projects/gcp-project1/logs/cloudaudit.googleapis.com%2Factivity",
            "p_any_ip_addresses": ["1.2.3.4"],
            "p_event_time": "2023-03-09 15:49:21.715",
            "p_log_type": "GCP.AuditLog",
            "p_parse_time": "2023-03-09 15:50:14.612",
            "p_row_id": "9a08afb7bb84d9eef3d98cee16b306",
            "p_source_id": "4fc88a5a-2d51-4279-9c4a-08fa7cc52566",
            "p_source_label": "gcplogsource",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {
                    "principalEmail": "staging@company.io",
                    "principalSubject": "user:staging@company.io",
                },
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "iam.serviceAccounts.create",
                        "resource": "projects/gcp-project1",
                        "resourceAttributes": {},
                    },
                ],
                "methodName": "google.iam.admin.v1.CreateServiceAccount",
                "request": {
                    "@type": "type.googleapis.com/google.iam.admin.v1.CreateServiceAccountRequest",
                    "account_id": "created-service-account",
                    "name": "projects/gcp-project1",
                    "service_account": {"description": "sa created", "display_name": "created-service-account"},
                },
                "requestMetadata": {
                    "callerIP": "1.2.3.4",
                    "callerSuppliedUserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36,gzip(gfe)",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2023-03-09T15:49:21.729005262Z"},
                },
                "resourceName": "projects/gcp-project1",
                "response": {
                    "@type": "type.googleapis.com/google.iam.admin.v1.ServiceAccount",
                    "description": "sa with limited permissions",
                    "display_name": "created-service-account",
                    "email": "created-service-account@gcp-project1.iam.gserviceaccount.com",
                    "etag": "MDEwMjE5MjA=",
                    "name": "projects/gcp-project1/serviceAccounts/created-service-account@gcp-project1.iam.gserviceaccount.com",
                    "oauth2_client_id": "103995727391704507722",
                    "project_id": "gcp-project1",
                    "unique_id": "103995727391704507722",
                },
                "serviceName": "iam.googleapis.com",
                "status": {},
            },
            "receiveTimestamp": "2023-03-09 15:49:22.599",
            "resource": {
                "labels": {
                    "email_id": "created-service-account@gcp-project1.iam.gserviceaccount.com",
                    "project_id": "gcp-project1",
                    "unique_id": "103995727391704507722",
                },
                "type": "service_account",
            },
            "severity": "NOTICE",
            "timestamp": "2023-03-09 15:49:21.715",
        },
    ),
    RuleTest(
        name="Other",
        expected_result=False,
        log={
            "insertId": "46ee5sd38mw",
            "logName": "projects/gcp-project1/logs/cloudaudit.googleapis.com%2Factivity",
            "p_any_emails": ["staging@company.io"],
            "p_any_ip_addresses": ["1.2.3.4"],
            "p_event_time": "2023-04-25 19:20:57.024",
            "p_log_type": "GCP.AuditLog",
            "p_parse_time": "2023-04-25 19:22:14.743",
            "p_row_id": "b2e9b7f5dc85a69981fac2e417b6bb03",
            "p_schema_version": 0,
            "p_source_id": "5b77391b-afad-46c7-8ddc-b8e21d4726b3",
            "p_source_label": "gcplogsource2",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {"principalEmail": "staging@company.io"},
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "iap.webServices.setIamPolicy",
                        "resourceAttributes": {
                            "name": "projects/123456789012/iap_web/compute/services/7312383563505470445",
                            "service": "iap.googleapis.com",
                            "type": "iap.googleapis.com/WebService",
                        },
                    },
                ],
                "methodName": "google.cloud.iap.v1.IdentityAwareProxyAdminService.SetIamPolicy",
                "request": {
                    "@type": "type.googleapis.com/google.iam.v1.SetIamPolicyRequest",
                    "policy": {"etag": "BwX6LgT4YMw="},
                    "resource": "projects/123456789012/iap_web/compute/services/7312383563505470445",
                },
                "requestMetadata": {
                    "callerIP": "1.2.3.4",
                    "callerSuppliedUserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36,gzip(gfe),gzip(gfe)",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2023-04-25T19:20:57.295723118Z"},
                },
                "resourceName": "projects/123456789012/iap_web/compute/services/7312383563505470445",
                "response": {"@type": "type.googleapis.com/google.iam.v1.Policy", "etag": "BwX6LgXbpsw="},
                "serviceName": "iap.googleapis.com",
            },
            "receiveTimestamp": "2023-04-25 19:20:58.16",
            "resource": {
                "labels": {"backend_service_id": "", "location": "", "project_id": "gcp-project1"},
                "type": "gce_backend_service",
            },
            "severity": "NOTICE",
            "timestamp": "2023-04-25 19:20:57.024",
        },
    ),
]


@panther_managed
class GCPServiceAccountorKeysCreated(Rule):
    default_description = (
        "Detects when a service account or key is created manually by a user instead of an automated workflow."
    )
    display_name = "GCP Service Account or Keys Created "
    default_reference = "https://cloud.google.com/iam/docs/keys-create-delete"
    default_severity = Severity.LOW
    log_types = [LogType.GCP_AUDIT_LOG]
    id = "GCP.Service.Account.or.Keys.Created-prototype"
    tests = gcp_service_accountor_keys_created_tests

    def rule(self, event):
        return all(
            [
                deep_get(event, "resource", "type", default="") == "service_account",
                "CreateServiceAccount" in deep_get(event, "protoPayload", "methodName", default=""),
                not deep_get(event, "protoPayload", "authenticationInfo", "principalEmail", default="").endswith(
                    ".gserviceaccount.com",
                ),
            ],
        )

    def title(self, event):
        actor = deep_get(event, "protoPayload", "authenticationInfo", "principalEmail", default="<ACTOR_NOT_FOUND>")
        target = deep_get(event, "resource", "labels", "email_id")
        project = deep_get(event, "resource", "labels", "project_id")
        resource = (
            "Service Account Key for"
            if deep_get(event, "protoPayload", "methodName", default="")
            == "google.iam.admin.v1.CreateServiceAccountKey"
            else "Service Account"
        )
        return f"GCP: [{actor}] created {resource} [{target}] in project [{project}]"



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get, okta_alert_context

okta_threat_insight_security_threat_detected_tests: list[RuleTest] = [
    RuleTest(
        name="Other Event",
        expected_result=False,
        log={
            "actor": {
                "alternateId": "homer.simpson@duff.com",
                "displayName": "Homer Simpson",
                "id": "00abc456",
                "type": "User",
            },
            "authenticationcontext": {"authenticationStep": 0, "externalSessionId": "abc12345"},
            "client": {
                "device": "Unknown",
                "ipAddress": "1.2.3.4",
                "userAgent": {"browser": "UNKNOWN", "os": "Unknown", "rawUserAgent": "Chrome"},
                "zone": "null",
            },
            "debugcontext": {"debugData": {}},
            "eventtype": "application.integration.rate_limit_exceeded",
            "legacyeventtype": "app.api.error.rate.limit.exceeded",
            "outcome": {"result": "SUCCESS"},
            "published": "2022-06-10 17:19:58.423",
            "request": {},
            "securitycontext": {},
            "severity": "INFO",
            "target": [{"alternateId": "App ", "displayName": "App", "id": "12345", "type": "AppInstance"}],
            "transaction": {"detail": {}, "id": "sdfg", "type": "JOB"},
            "uuid": "aaa-bb-ccc",
            "version": "0",
        },
    ),
    RuleTest(
        name="Threat Detected Event - Deny",
        expected_result=True,
        log={
            "actor": {"alternateId": "unknown", "displayName": "1.2.3.4", "id": "1.2.3.4", "type": "IP address"},
            "authenticationcontext": {"authenticationStep": 0},
            "client": {
                "device": "Computer",
                "geographicalContext": {
                    "city": "Dallas",
                    "country": "United States",
                    "geolocation": {"lat": 32.7908, "lon": -96.8336},
                    "postalCode": "75207",
                    "state": "Texas",
                },
                "ipAddress": "1.2.3.4",
                "userAgent": {
                    "browser": "CHROME",
                    "os": "Windows 10",
                    "rawUserAgent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
                },
                "zone": "null",
            },
            "debugcontext": {
                "debugData": {
                    "dtHash": "abcsadfjsald",
                    "requestId": "alsdjflasf",
                    "requestUri": "/oauth2/v1/authorize",
                    "threatDetections": '{"Login failures with high unknown users count":"HIGH","Password Spray":"HIGH","Login Failures":"MEDIUM"}',
                    "threatSuspected": "true",
                    "url": "/oauth2/v1/authorize",
                },
            },
            "displaymessage": "Request from suspicious actor",
            "eventtype": "security.threat.detected",
            "legacyeventtype": "security.threat.detected",
            "outcome": {
                "reason": "Password Spray, Login failures with high unknown users count, Login Failures",
                "result": "DENY",
            },
            "published": "2022-12-14 19:16:32.015",
            "request": {
                "ipChain": [
                    {
                        "geographicalContext": {
                            "city": "Dallas",
                            "country": "United States",
                            "geolocation": {"lat": 32.7908, "lon": -96.8336},
                            "postalCode": "75207",
                            "state": "Texas",
                            "ip": "1.2.3.4",
                            "version": "V4",
                        },
                    },
                ],
            },
            "securitycontext": {
                "asNumber": 62240,
                "asOrg": "packethub s.a.",
                "domain": ".",
                "isProxy": False,
                "isp": "clouvider limited",
            },
            "severity": "WARN",
            "transaction": {"detail": {}, "id": "asdfjaslf", "type": "WEB"},
            "uuid": "asdfa-1234-asdfdas",
            "version": "0",
        },
    ),
    RuleTest(
        name="Threat Detected Event - Success",
        expected_result=True,
        log={
            "actor": {"alternateId": "unknown", "displayName": "1.2.3.4", "id": "1.2.3.4", "type": "IP address"},
            "authenticationcontext": {"authenticationStep": 0},
            "client": {
                "device": "Computer",
                "userAgent": {
                    "browser": "CHROME",
                    "os": "Windows 10",
                    "rawUserAgent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
                },
                "zone": "null",
            },
            "debugcontext": {
                "debugData": {
                    "dtHash": "abcsadfjsald",
                    "requestId": "alsdjflasf",
                    "requestUri": "/oauth2/v1/authorize",
                    "threatDetections": '{"Login failures with high unknown users count":"HIGH","Password Spray":"HIGH","Login Failures":"MEDIUM"}',
                    "threatSuspected": "true",
                    "url": "/oauth2/v1/authorize",
                },
            },
            "displaymessage": "Request from suspicious actor",
            "eventtype": "security.threat.detected",
            "legacyeventtype": "security.threat.detected",
            "outcome": {
                "reason": "Password Spray, Login failures with high unknown users count, Login Failures",
                "result": "SUCCESS",
            },
            "published": "2022-12-14 19:16:32.015",
            "request": {
                "ipChain": [
                    {
                        "geographicalContext": {
                            "city": "Dallas",
                            "country": "United States",
                            "geolocation": {"lat": 32.7908, "lon": -96.8336},
                            "postalCode": "75207",
                            "state": "Texas",
                            "ip": "1.2.3.4",
                            "version": "V4",
                        },
                    },
                ],
            },
            "securitycontext": {
                "asNumber": 62240,
                "asOrg": "packethub s.a.",
                "domain": ".",
                "isProxy": False,
                "isp": "clouvider limited",
            },
            "severity": "WARN",
            "transaction": {"detail": {}, "id": "asdfjaslf", "type": "WEB"},
            "uuid": "asdfa-1234-asdfdas",
            "version": "0",
        },
    ),
]


@panther_managed
class OktaThreatInsightSecurityThreatDetected(Rule):
    default_description = "Okta ThreatInsight identified request from potentially malicious IP address"
    default_reference = (
        "https://help.okta.com/en-us/Content/Topics/Security/threat-insight/configure-threatinsight-system-log.htm"
    )
    display_name = "Okta ThreatInsight Security Threat Detected"
    default_severity = Severity.HIGH
    log_types = [LogType.OKTA_SYSTEM_LOG]
    id = "Okta.ThreatInsight.Security.Threat.Detected-prototype"
    tests = okta_threat_insight_security_threat_detected_tests

    def severity_from_threat_string(self, threat_detection):
        # threat detection is a string but contains json data
        # can contain multiple threats detected with multiple severities
        # return highest found severity
        if "CRITICAL" in threat_detection:
            return "CRITICAL"
        if "HIGH" in threat_detection:
            return "HIGH"
        if "MEDIUM" in threat_detection:
            return "MEDIUM"
        if "LOW" in threat_detection:
            return "LOW"
        if "INFO" in threat_detection:
            return "INFO"
        return "MEDIUM"

    def rule(self, event):
        return event.get("eventtype") == "security.threat.detected"

    def title(self, event):
        return f"Okta: ThreatInsight identified potentially malicious behavior for [{event.get('actor', {}).get('displayName', '<display-name-not-found>')}]"

    def severity(self, event):
        outcome = deep_get(event, "outcome", "result", default="<OUTCOME_NOT_FOUND>")
        if outcome == "DENY":
            return "INFO"
        threat_detection = (
            event.get("debugcontext", {}).get("debugData", {}).get("threatDetections", "<threat-detection-not-found>")
        )
        return self.severity_from_threat_string(threat_detection)

    def alert_context(self, event):
        return okta_alert_context(event)



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get
from pypanther.helpers.panther_snyk_helpers import snyk_alert_context

snyk_ou_change_tests: list[RuleTest] = [
    RuleTest(
        name="Snyk Org Deletion ( HIGH )",
        expected_result=True,
        log={
            "groupId": "8fffffff-1555-4444-b000-b55555555555",
            "orgId": "21111111-a222-4eee-8ddd-a99999999999",
            "event": "org.delete",
            "content": {"orgName": "expendable-org"},
            "created": "2023-04-09T23:32:14.649Z",
            "userId": "05555555-3333-4ddd-8ccc-755555555555",
        },
    ),
    RuleTest(
        name="Snyk Group Org Remove ( HIGH )",
        expected_result=True,
        log={
            "groupId": "8fffffff-1555-4444-b000-b55555555555",
            "orgId": "21111111-a222-4eee-8ddd-a99999999999",
            "event": "group.org.remove",
            "content": {"orgName": "expendable-org"},
            "created": "2023-04-09T23:32:14.649Z",
            "userId": "05555555-3333-4ddd-8ccc-755555555555",
        },
    ),
    RuleTest(
        name="Snyk Group Edit ( MEDIUM )",
        expected_result=True,
        log={
            "groupId": "8fffffff-1555-4444-b000-b55555555555",
            "event": "group.edit",
            "content": {"updatedValues": {"projectTestFrequencySetting": "daily"}},
            "created": "2023-04-11T23:22:57.667Z",
        },
    ),
    RuleTest(
        name="Snyk Org Create ( INFO )",
        expected_result=True,
        log={
            "groupId": "8fffffff-1555-4444-b000-b55555555555",
            "event": "org.create",
            "content": {"newOrgPublicId": "21111111-a222-4eee-8ddd-a99999999999"},
            "created": "2023-04-11T23:12:33.206Z",
            "userId": "05555555-3333-4ddd-8ccc-755555555555",
        },
    ),
    RuleTest(
        name="Snyk Group SSO Membership sync",
        expected_result=False,
        log={
            "content": {},
            "created": "2023-03-15 13:13:13.133",
            "event": "group.sso.membership.sync",
            "groupId": "8fffffff-1555-4444-b000-b55555555555",
        },
    ),
]


@panther_managed
class SnykOUChange(Rule):
    id = "Snyk.OU.Change-prototype"
    display_name = "Snyk Org or Group Settings Change"
    log_types = [LogType.SNYK_GROUP_AUDIT, LogType.SNYK_ORG_AUDIT]
    tags = ["Snyk"]
    default_severity = Severity.HIGH
    default_description = "Detects when Snyk Group or Organization Settings are changed.\n"
    default_runbook = "These actions in the Snyk Audit logs indicate that a Organization or Group setting has changed, including Group and Org creation/deletion. Deletion events are marked with HIGH severity Creation events are marked with INFO severity Edit events are marked with MEDIUM Severity\n"
    default_reference = "https://docs.snyk.io/snyk-admin/introduction-to-snyk-administration"
    summary_attributes = ["event"]
    tests = snyk_ou_change_tests
    ACTIONS = [
        "group.create",
        "group.delete",
        "group.edit",
        "group.feature_flags.edit",
        "group.org.add",
        "group.org.remove",
        "group.settings.edit",
        "group.settings.feature_flag.edit",
        "org.create",
        "org.delete",
        "org.edit",
        "org.settings.feature_flag.edit",
    ]

    def rule(self, event):
        action = deep_get(event, "event", default="<NO_EVENT>")
        return action in self.ACTIONS

    def title(self, event):
        group_or_org = "<GROUP_OR_ORG>"
        action = deep_get(event, "event", default="<NO_EVENT>")
        if "." in action:
            group_or_org = action.split(".")[0].title()
        return f"Snyk: [{group_or_org}] Organizational Unit settings have been modified via [{action}] performed by [{deep_get(event, 'userId', default='<NO_USERID>')}]"

    def alert_context(self, event):
        return snyk_alert_context(event)

    def dedup(self, event):
        return f"{deep_get(event, 'userId', default='<NO_USERID>')}{deep_get(event, 'orgId', default='<NO_ORGID>')}{deep_get(event, 'groupId', default='<NO_GROUPID>')}{deep_get(event, 'event', default='<NO_EVENT>')}"

    def severity(self, event):
        action = deep_get(event, "event", default="<NO_EVENT>")
        if action.endswith((".remove", ".delete")):
            return "HIGH"
        if action.endswith(".edit"):
            return "MEDIUM"
        return "INFO"



import pypanther.helpers.panther_event_type_helpers as event_type
from pypanther.base import DataModel, DataModelMapping, LogType

ADMIN_EVENTS = {
    "business.add_admin",
    "business.invite_admin",
    "team.promote_maintainer",
}

CONDITIONAL_ADMIN_EVENTS = {
    "team.add_repository",
}


def get_admin_role(event):
    action = event.get("action", "")
    permission = event.get("permission", "")
    if action in CONDITIONAL_ADMIN_EVENTS and permission == "admin":
        return action
    return action if action in ADMIN_EVENTS else "<UNKNOWN_ADMIN_ROLE>"


def get_event_type(event):
    action = event.get("action", "")
    permission = event.get("permission", "")
    if action in ADMIN_EVENTS:
        return event_type.ADMIN_ROLE_ASSIGNED
    if action in CONDITIONAL_ADMIN_EVENTS and permission == "admin":
        return event_type.ADMIN_ROLE_ASSIGNED
    if event.get("action", "") == "org.disable_two_factor_requirement":
        return event_type.MFA_DISABLED
    return None


class StandardGithubAudit(DataModel):
    id: str = "Standard.Github.Audit"
    display_name: str = "Github Audit"
    enabled: bool = True
    log_types: list[str] = [LogType.GITHUB_AUDIT]
    mappings: list[DataModelMapping] = [
        DataModelMapping(name="actor_user", path="actor"),
        DataModelMapping(name="assigned_admin_role", method=get_admin_role),
        DataModelMapping(name="event_type", method=get_event_type),
        DataModelMapping(name="user", path="user"),
    ]



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get

box_content_workflow_policy_violation_tests: list[RuleTest] = [
    RuleTest(
        name="Regular Event",
        expected_result=False,
        log={
            "type": "event",
            "additional_details": '{"key": "value"}',
            "created_by": {"id": "12345678", "type": "user", "login": "cat@example", "name": "Bob Cat"},
            "event_type": "DELETE",
        },
    ),
    RuleTest(
        name="Upload Policy Violation",
        expected_result=True,
        log={
            "type": "event",
            "additional_details": '{"key": "value"}',
            "created_by": {"id": "12345678", "type": "user", "login": "cat@example", "name": "Bob Cat"},
            "event_type": "CONTENT_WORKFLOW_UPLOAD_POLICY_VIOLATION",
            "source": {"id": "12345678", "type": "user", "login": "user@example"},
        },
    ),
    RuleTest(
        name="Sharing Policy Violation",
        expected_result=True,
        log={
            "type": "event",
            "additional_details": {"key": "value"},
            "created_by": {"id": "12345678", "type": "user", "login": "cat@example", "name": "Mountain Lion"},
            "event_type": "CONTENT_WORKFLOW_SHARING_POLICY_VIOLATION",
            "source": {"id": "12345678", "type": "user", "login": "user@example"},
        },
    ),
]


@panther_managed
class BoxContentWorkflowPolicyViolation(Rule):
    id = "Box.Content.Workflow.Policy.Violation-prototype"
    display_name = "Box Content Workflow Policy Violation"
    log_types = [LogType.BOX_EVENT]
    tags = ["Box"]
    default_severity = Severity.LOW
    default_description = "A user violated the content workflow policy.\n"
    default_reference = "https://support.box.com/hc/en-us/articles/360043692594-Creating-a-Security-Policy"
    default_runbook = "Investigate whether the user continues to violate the policy and take measure to ensure they understand policy.\n"
    summary_attributes = ["event_type"]
    tests = box_content_workflow_policy_violation_tests
    POLICY_VIOLATIONS = {"CONTENT_WORKFLOW_UPLOAD_POLICY_VIOLATION", "CONTENT_WORKFLOW_SHARING_POLICY_VIOLATION"}

    def rule(self, event):
        return event.get("event_type") in self.POLICY_VIOLATIONS

    def title(self, event):
        return f"User [{deep_get(event, 'created_by', 'name', default='<UNKNOWN_USER>')}] violated a content workflow policy."



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get, deep_walk

gcp_permissions_grantedto_createor_manage_service_account_key_tests: list[RuleTest] = [
    RuleTest(
        name="other event",
        expected_result=False,
        log={
            "insertid": "abcdefghijklmn",
            "logname": "projects/gcp-project1/logs/cloudaudit.googleapis.com%2Factivity",
            "operation": {
                "id": "1234567890123-gcp-project1:abcdefghijklmnopqrstuvwz",
                "last": True,
                "producer": "bigquery.googleapis.com",
            },
            "p_any_emails": ["user@company.io"],
            "p_any_ip_addresses": ["1.2.3.4"],
            "p_event_time": "2023-03-28 18:37:06.079",
            "p_log_type": "GCP.AuditLog",
            "p_parse_time": "2023-03-28 18:38:14.478",
            "p_row_id": "06bf03d9d5dfbadba981899e1787bf05",
            "p_schema_version": 0,
            "p_source_id": "964c7894-9a0d-4ddf-864f-0193438221d6",
            "p_source_label": "gcp-logsource",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {"principalEmail": "user@company.io"},
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "bigquery.tables.delete",
                        "resource": "projects/gcp-project1/datasets/test1/tables/newtable",
                    },
                ],
                "metadata": {
                    "@type": "type.googleapis.com/google.cloud.audit.BigQueryAuditMetadata",
                    "methodName": "google.cloud.bigquery.v2.JobService.InsertJob",
                    "requestMetadata": {
                        "callerIP": "1.2.3.4",
                        "callerSuppliedUserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36,gzip(gfe),gzip(gfe)",
                    },
                    "resourceName": "projects/gcp-project1/datasets/test1/tables/newtable",
                    "serviceName": "bigquery.googleapis.com",
                    "status": {},
                    "tableDeletion": {
                        "jobName": "projects/gcp-project1/jobs/bquxjob_5e4a0679_18729a639d7",
                        "reason": "QUERY",
                    },
                },
                "receivetimestamp": "2023-03-28 18:37:06.745",
                "resource": {
                    "labels": {"dataset_id": "test1", "project_id": "gcp-project1"},
                    "type": "bigquery_dataset",
                },
                "severity": "NOTICE",
                "timestamp": "2023-03-28 18:37:06.079",
            },
        },
    ),
    RuleTest(
        name="service account match",
        expected_result=True,
        log={
            "insertId": "hhpfjvdgakc",
            "logName": "projects/gcp-project1/logs/cloudaudit.googleapis.com%2Factivity",
            "p_any_emails": ["user@company.io"],
            "p_any_ip_addresses": ["1.2.3.4"],
            "p_event_time": "2023-04-10 18:36:30.838",
            "p_log_type": "GCP.AuditLog",
            "p_parse_time": "2023-04-10 18:38:14.607",
            "p_row_id": "5286b52d4095c9f1b2e8eabe178f8203",
            "p_schema_version": 0,
            "p_source_id": "5b77391b-afad-46c7-8ddc-b8e21d4726b3",
            "p_source_label": "gcplogsource2",
            "protoPayload": {
                "at_sign_type": "type.googleapis.com/google.cloud.audit.AuditLog",
                "authenticationInfo": {"principalEmail": "user@company.io", "principalSubject": "user:user@company.io"},
                "authorizationInfo": [
                    {
                        "granted": True,
                        "permission": "iam.serviceAccounts.setIamPolicy",
                        "resource": "projects/-/serviceAccounts/105537103139416651075",
                        "resourceAttributes": {"name": "projects/-/serviceAccounts/105537103139416651075"},
                    },
                ],
                "methodName": "google.iam.admin.v1.SetIAMPolicy",
                "request": {
                    "@type": "type.googleapis.com/google.iam.v1.SetIamPolicyRequest",
                    "policy": {
                        "bindings": [
                            {
                                "members": ["serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com"],
                                "role": "roles/iam.serviceAccountTokenCreator",
                            },
                            {
                                "members": ["serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com"],
                                "role": "roles/iam.serviceAccountUser",
                            },
                        ],
                        "etag": "ACAB",
                        "version": 3,
                    },
                    "resource": "projects/gcp-project1/serviceAccounts/105537103139416651075",
                },
                "requestMetadata": {
                    "callerIP": "1.2.3.4",
                    "callerSuppliedUserAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36,gzip(gfe)",
                    "destinationAttributes": {},
                    "requestAttributes": {"auth": {}, "time": "2023-04-10T18:36:30.994141642Z"},
                },
                "resourceName": "projects/-/serviceAccounts/105537103139416651075",
                "response": {
                    "@type": "type.googleapis.com/google.iam.v1.Policy",
                    "bindings": [
                        {
                            "members": ["serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com"],
                            "role": "roles/iam.serviceAccountTokenCreator",
                        },
                        {
                            "members": ["serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com"],
                            "role": "roles/iam.serviceAccountUser",
                        },
                    ],
                    "etag": "BwX4/6dQjX4=",
                    "version": 1,
                },
                "serviceData": {
                    "@type": "type.googleapis.com/google.iam.v1.logging.AuditData",
                    "policyDelta": {
                        "bindingDeltas": [
                            {
                                "action": "ADD",
                                "member": "serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com",
                                "role": "roles/iam.serviceAccountTokenCreator",
                            },
                            {
                                "action": "ADD",
                                "member": "serviceAccount:test-account3@gcp-project1.iam.gserviceaccount.com",
                                "role": "roles/iam.serviceAccountUser",
                            },
                        ],
                    },
                },
                "serviceName": "iam.googleapis.com",
                "status": {},
            },
            "receiveTimestamp": "2023-04-10 18:36:32.268",
            "resource": {
                "labels": {
                    "email_id": "test-account3@gcp-project1.iam.gserviceaccount.com",
                    "project_id": "gcp-project1",
                    "unique_id": "105537103139416651075",
                },
                "type": "service_account",
            },
            "severity": "NOTICE",
            "timestamp": "2023-04-10 18:36:30.838",
        },
    ),
]


@panther_managed
class GCPPermissionsGrantedtoCreateorManageServiceAccountKey(Rule):
    default_description = "Permissions granted to impersonate a service account. This includes predefined service account IAM roles granted at the parent project, folder or organization-level."
    display_name = "GCP Permissions Granted to Create or Manage Service Account Key"
    default_reference = "https://cloud.google.com/iam/docs/keys-create-delete"
    default_severity = Severity.LOW
    log_types = [LogType.GCP_AUDIT_LOG]
    id = "GCP.Permissions.Granted.to.Create.or.Manage.Service.Account.Key-prototype"
    tests = gcp_permissions_grantedto_createor_manage_service_account_key_tests
    SERVICE_ACCOUNT_MANAGE_ROLES = ["roles/iam.serviceAccountTokenCreator", "roles/iam.serviceAccountUser"]

    def rule(self, event):
        if "SetIAMPolicy" in deep_get(event, "protoPayload", "methodName", default=""):
            role = deep_walk(
                event,
                "ProtoPayload",
                "serviceData",
                "policyDelta",
                "bindingDeltas",
                "role",
                default="",
                return_val="last",
            )
            action = deep_walk(
                event,
                "ProtoPayload",
                "serviceData",
                "policyDelta",
                "bindingDeltas",
                "action",
                default="",
                return_val="last",
            )
            return role in self.SERVICE_ACCOUNT_MANAGE_ROLES and action == "ADD"
        return False

    def title(self, event):
        actor = deep_get(event, "protoPayload", "authenticationInfo", "principalEmail", default="<ACTOR_NOT_FOUND>")
        target = deep_get(event, "resource", "labels", "email_id") or deep_get(
            event,
            "resource",
            "labels",
            "project_id",
            default="<TARGET_NOT_FOUND>",
        )
        return f"GCP: [{actor}] granted permissions to create or manage service account keys to [{target}]"

    def alert_context(self, event):
        return {"resource": deep_get(event, "resource"), "serviceData": deep_get(event, "protoPayload", "serviceData")}



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import github_alert_context

git_hub_org_moderators_add_tests: list[RuleTest] = [
    RuleTest(
        name="GitHub - Org Moderator Added",
        expected_result=True,
        log={
            "_document_id": "Ab123",
            "action": "organization_moderators.add_user",
            "actor": "sarah78",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-11 05:17:28.078",
            "created_at": "2022-12-11 05:17:28.078",
            "org": "example-io",
            "user": "john1987",
        },
    ),
    RuleTest(
        name="GitHub - Org Moderator removed",
        expected_result=False,
        log={
            "_document_id": "Ab123",
            "action": "organization_moderators.remove_user",
            "actor": "sarah78",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-11 05:17:28.078",
            "created_at": "2022-12-11 05:17:28.078",
            "org": "example-io",
            "user": "john1987",
        },
    ),
]


@panther_managed
class GitHubOrgModeratorsAdd(Rule):
    id = "GitHub.Org.Moderators.Add-prototype"
    display_name = "GitHub User Added to Org Moderators"
    log_types = [LogType.GITHUB_AUDIT]
    tags = ["GitHub", "Initial Access:Supply Chain Compromise"]
    default_severity = Severity.MEDIUM
    default_description = "Detects when a user is added to a GitHub org's list of moderators."
    default_reference = "https://docs.github.com/en/organizations/managing-peoples-access-to-your-organization-with-roles/managing-moderators-in-your-organization"
    tests = git_hub_org_moderators_add_tests

    def rule(self, event):
        return event.get("action") == "organization_moderators.add_user"

    def title(self, event):
        return f"GitHub.Audit: User [{event.get('actor', '<UNKNOWN_ACTOR>')}] added user [{event.get('user', '<UNKNOWN_USER>')}] to moderators in [{event.get('org', '<UNKNOWN_ORG>')}]"

    def alert_context(self, event):
        return github_alert_context(event)



import shlex

from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import deep_get

osquery_linux_aws_command_executed_tests: list[RuleTest] = [
    RuleTest(
        name="AWS command executed on MacOS",
        expected_result=False,
        log={
            "name": "pack_incident-response_shell_history",
            "action": "added",
            "decorations": {"platform": "darwin"},
            "columns": {
                "command": "aws sts get-caller-identity",
                "uid": "1000",
                "directory": "/home/ubuntu",
                "username": "ubuntu",
            },
        },
    ),
    RuleTest(
        name="AWS command executed",
        expected_result=True,
        log={
            "name": "pack_incident-response_shell_history",
            "action": "added",
            "columns": {"command": "aws s3 ls", "uid": "1000", "directory": "/home/ubuntu", "username": "ubuntu"},
        },
    ),
    RuleTest(
        name="Tail command executed",
        expected_result=False,
        log={
            "name": "pack_incident-response_shell_history",
            "action": "added",
            "columns": {
                "command": "tail -f /var/log/all",
                "uid": "1000",
                "directory": "/home/ubuntu",
                "username": "ubuntu",
            },
        },
    ),
    RuleTest(
        name="Command with quote executed",
        expected_result=False,
        log={
            "name": "pack_incident-response_shell_history",
            "action": "added",
            "columns": {
                "command": "git commit -m 'all done'",
                "uid": "1000",
                "directory": "/home/ubuntu",
                "username": "ubuntu",
            },
        },
    ),
    RuleTest(
        name="Invalid command ignored",
        expected_result=False,
        log={
            "name": "pack_incident-response_shell_history",
            "action": "added",
            "columns": {"command": "unopened '", "uid": "1000", "directory": "/home/ubuntu", "username": "ubuntu"},
        },
    ),
]


@panther_managed
class OsqueryLinuxAWSCommandExecuted(Rule):
    id = "Osquery.Linux.AWSCommandExecuted-prototype"
    display_name = "AWS command executed on the command line"
    log_types = [LogType.OSQUERY_DIFFERENTIAL]
    tags = ["Osquery", "Linux", "Execution:User Execution"]
    reports = {"MITRE ATT&CK": ["TA0002:T1204"]}
    default_severity = Severity.MEDIUM
    default_description = "An AWS command was executed on a Linux instance"
    default_runbook = "See which other commands were executed, and then remove IAM role causing the access"
    default_reference = "https://attack.mitre.org/techniques/T1078/"
    summary_attributes = ["name", "action"]
    tests = osquery_linux_aws_command_executed_tests
    PLATFORM_IGNORE_LIST = {"darwin"}

    def rule(self, event):
        # Filter out irrelevant logs & systems
        if (
            event.get("action") != "added"
            or "shell_history" not in event.get("name")
            or deep_get(event, "decorations", "platform") in self.PLATFORM_IGNORE_LIST
        ):
            return False
        command = deep_get(event, "columns", "command")
        if not command:
            return False
        try:
            command_args = shlex.split(command)
        except ValueError:
            # "No escaped character" or "No closing quotation", probably an invalid command
            return False
        if command_args[0] == "aws":
            return True
        return False

    def title(self, event):
        return f"User [{deep_get(event, 'columns', 'username', default='<UNKNOWN_USER>')}] issued an aws-cli command on [{event.get('hostIdentifier', '<UNKNOWN_HOST>')}]"



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.crowdstrike_event_streams_helpers import cs_alert_context

crowdstrike_api_key_deleted_tests: list[RuleTest] = [
    RuleTest(
        name="API Key Deleted",
        expected_result=True,
        log={
            "event": {
                "AuditKeyValues": [
                    {"Key": "scope(s)", "ValueString": "alerts:read,api-integrations:read"},
                    {"Key": "actor_user", "ValueString": "tester@panther.com"},
                    {"Key": "actor_user_uuid", "ValueString": "a11a1111-1a11-1a1a-1a11-a11a111a111a"},
                    {"Key": "actor_cid", "ValueString": "aaa111111111111111aaaaaa11a11a11"},
                    {"Key": "trace_id", "ValueString": "1a111111-a1a1-111a-11aa-a111111a1a1a"},
                    {"Key": "APIClientID", "ValueString": "aaa1a11aaa111a1a11a11aaaa1aa1a11"},
                    {"Key": "id", "ValueString": "aaa1a11aaa111a1a11a11aaaa1aa1a11"},
                    {"Key": "name", "ValueString": "key name"},
                ],
                "OperationName": "DeleteAPIClients",
                "ServiceName": "Crowdstrike API Client",
                "Success": True,
                "UTCTimestamp": "2024-07-08 14:01:54.000000000",
                "UserId": "tester@panther.com",
                "UserIp": "11.1.111.11",
            },
            "metadata": {
                "customerIDString": "aaa111111111111111aaaaaa11a11a11",
                "eventCreationTime": "2024-07-08 14:01:54.451000000",
                "eventType": "AuthActivityAuditEvent",
                "offset": 111111,
                "version": "1.0",
            },
        },
    ),
    RuleTest(
        name="API Key Delete failed",
        expected_result=False,
        log={
            "event": {
                "AuditKeyValues": [
                    {"Key": "scope(s)", "ValueString": "alerts:read,api-integrations:read"},
                    {"Key": "actor_user", "ValueString": "tester@panther.com"},
                    {"Key": "actor_user_uuid", "ValueString": "a11a1111-1a11-1a1a-1a11-a11a111a111a"},
                    {"Key": "actor_cid", "ValueString": "aaa111111111111111aaaaaa11a11a11"},
                    {"Key": "trace_id", "ValueString": "1a111111-a1a1-111a-11aa-a111111a1a1a"},
                    {"Key": "APIClientID", "ValueString": "aaa1a11aaa111a1a11a11aaaa1aa1a11"},
                    {"Key": "id", "ValueString": "aaa1a11aaa111a1a11a11aaaa1aa1a11"},
                    {"Key": "name", "ValueString": "key name"},
                ],
                "OperationName": "DeleteAPIClients",
                "ServiceName": "Crowdstrike API Client",
                "Success": False,
                "UTCTimestamp": "2024-07-08 14:01:54.000000000",
                "UserId": "tester@panther.com",
                "UserIp": "11.1.111.11",
            },
            "metadata": {
                "customerIDString": "aaa111111111111111aaaaaa11a11a11",
                "eventCreationTime": "2024-07-08 14:01:54.451000000",
                "eventType": "AuthActivityAuditEvent",
                "offset": 111111,
                "version": "1.0",
            },
        },
    ),
]


@panther_managed
class CrowdstrikeAPIKeyDeleted(Rule):
    id = "Crowdstrike.API.Key.Deleted-prototype"
    display_name = "Crowdstrike API Key Deleted"
    reports = {"MITRE ATT&CK": ["T1531", "T1070"]}
    log_types = [LogType.CROWDSTRIKE_EVENT_STREAMS]
    default_severity = Severity.MEDIUM
    default_description = "A user deleted an API Key in CrowdStrike"
    default_runbook = "Validate this action was authorized."
    tests = crowdstrike_api_key_deleted_tests

    def rule(self, event):
        return all([event.deep_get("event", "OperationName") == "DeleteAPIClients", event.deep_get("event", "Success")])

    def title(self, event):
        user = event.deep_get("event", "UserId")
        service = event.deep_get("event", "ServiceName")
        return f"{user} deleted an API key in {service}"

    def alert_context(self, event):
        return cs_alert_context(event)



import abc
import contextlib
import copy
import json
import logging
import os
import sys
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Literal, Mapping, Optional, Tuple, Type, Union
from unittest.mock import MagicMock, patch

from jsonpath_ng import Fields
from jsonpath_ng.ext import parse
from panther_core.detection import DetectionResult
from panther_core.enriched_event import PantherEvent
from panther_core.exceptions import FunctionReturnTypeError, UnknownDestinationError
from panther_core.util import get_bool_env_var
from pydantic import BaseModel, NonNegativeInt, PositiveInt, TypeAdapter

from pypanther.log_types import LogType
from pypanther.severity import SEVERITY_DEFAULT, SEVERITY_TYPES, Severity
from pypanther.unit_tests import RuleTest, RuleTestResult
from pypanther.utils import truncate, try_asdict
from pypanther.validate import NonEmptyUniqueList, UniqueList

logger = logging.getLogger(__name__)

# We want to default this to false as PAT will want detection output
DISABLE_OUTPUT = get_bool_env_var("DISABLE_DETECTION_OUTPUT", False)

TYPE_RULE = "RULE"
TYPE_SCHEDULED_RULE = "SCHEDULED_RULE"
TYPE_CORRELATION_RULE = "CORRELATION_RULE"

ERROR_TYPE_RULE = "RULE_ERROR"
ERROR_TYPE_SCHEDULED_RULE = "SCHEDULED_RULE_ERROR"
ERROR_TYPE_CORRELATION_RULE = "CORRELATION_RULE_ERROR"

# Maximum size for a dedup string
MAX_DEDUP_STRING_SIZE = 1000

# Maximum size for a generated field
MAX_GENERATED_FIELD_SIZE = 1000

# Maximum number of destinations
MAX_DESTINATIONS_SIZE = 10

# The limit for DDB is 400kb per item (we store this one in DDB) and
# the limit for SQS/SNS is 256KB. The limit of 200kb is an approximation - the other
# fields included in the request will be less than the remaining 56kb
MAX_ALERT_CONTEXT_SIZE = 200 * 1024  # 200kb

ALERT_CONTEXT_ERROR_KEY = "_error"

DEFAULT_DETECTION_DEDUP_PERIOD_MINS = 60

RULE_METHOD = "rule"

ALERT_CONTEXT_METHOD = "alert_context"
DEDUP_METHOD = "dedup"
DESCRIPTION_METHOD = "description"
DESTINATIONS_METHOD = "destinations"
REFERENCE_METHOD = "reference"
RUNBOOK_METHOD = "runbook"
SEVERITY_METHOD = "severity"
TITLE_METHOD = "title"

# Auxiliary METHODS are optional
AUXILIARY_METHODS = (
    ALERT_CONTEXT_METHOD,
    DEDUP_METHOD,
    DESCRIPTION_METHOD,
    DESTINATIONS_METHOD,
    REFERENCE_METHOD,
    RUNBOOK_METHOD,
    SEVERITY_METHOD,
    TITLE_METHOD,
)

RULE_ALL_METHODS = [
    RULE_METHOD,
    *AUXILIARY_METHODS,
]

RULE_ALL_ATTRS = [
    "create_alert",
    "dedup_period_minutes",
    "display_name",
    "enabled",
    "log_types",
    "id",
    "scheduled_queries",
    "summary_attributes",
    "tests",
    "threshold",
    "tags",
    "reports",
    "default_severity",
    "default_description",
    "default_destinations",
    "default_runbook",
    "default_reference",
]


@dataclass
class DataModelMapping:
    name: str
    path: Optional[str] = None
    method: Optional[Callable] = None


class DataModel:
    id: str
    display_name: str
    enabled: bool
    log_types: List[str]
    mappings: List[DataModelMapping]

    def __init__(self) -> None:
        self.paths: Dict[str, Fields] = {}
        self.methods: Dict[str, Callable] = {}

        for mapping in self.mappings:
            if not mapping.name:
                raise AssertionError(f"DataModel [{self.id}] is missing required field: [Name]")
            if mapping.path:
                self.paths[mapping.name] = parse(mapping.path)
            elif mapping.method:
                self.methods[mapping.name] = mapping.method
            else:
                raise AssertionError(f"DataModel [{self.id}] must define one of: [Path, Method]")


class RuleModel(BaseModel):
    create_alert: bool
    dedup_period_minutes: NonNegativeInt
    display_name: str
    enabled: bool
    log_types: NonEmptyUniqueList[str]
    id: str
    scheduled_queries: UniqueList[str]
    summary_attributes: UniqueList[str]
    tests: List[RuleTest]
    threshold: PositiveInt
    tags: UniqueList[str]
    reports: Dict[str, NonEmptyUniqueList[str]]
    default_destinations: UniqueList[str]
    default_description: str
    default_runbook: str
    default_reference: str
    default_severity: Severity


RuleAdapter = TypeAdapter(RuleModel)

DEFAULT_CREATE_ALERT = True
DEFAULT_DEDUP_PERIOD_MINUTES = 60
DEFAULT_DESCRIPTION = ""
DEFAULT_DISPLAY_NAME = ""
DEFAULT_ENABLED = True
DEFAULT_OUTPUT_IDS: List[str] = []
DEFAULT_REFERENCE = ""
DEFAULT_REPORTS: Dict[str, List[str]] = {}
DEFAULT_RUNBOOK = ""
DEFAULT_SCHEDULED_QUERIES: List[str] = []
DEFAULT_SUMMARY_ATTRIBUTES: List[str] = []
DEFAULT_TAGS: List[str] = []
DEFAULT_TESTS: List[RuleTest] = []
DEFAULT_THRESHOLD = 1

SeverityType = Union[Severity | Literal["DEFAULT"] | str]


class Rule(metaclass=abc.ABCMeta):
    """A Panther rule class. This class should be subclassed to create a new rule."""

    log_types: List[LogType | str]
    id: str
    create_alert: bool = DEFAULT_CREATE_ALERT
    dedup_period_minutes: NonNegativeInt = DEFAULT_DEDUP_PERIOD_MINUTES
    display_name: str = DEFAULT_DISPLAY_NAME
    enabled: bool = DEFAULT_ENABLED
    scheduled_queries: List[str] = DEFAULT_SCHEDULED_QUERIES
    summary_attributes: List[str] = DEFAULT_SUMMARY_ATTRIBUTES
    tests: List[RuleTest] = DEFAULT_TESTS
    threshold: PositiveInt = DEFAULT_THRESHOLD
    tags: List[str] = DEFAULT_TAGS
    reports: Dict[str, List[str]] = DEFAULT_REPORTS

    default_severity: Severity | str
    default_destinations: List[str] = DEFAULT_OUTPUT_IDS
    default_runbook: str = DEFAULT_RUNBOOK
    default_reference: str = DEFAULT_REFERENCE
    default_description: str = DEFAULT_DESCRIPTION

    def __str__(self) -> str:
        return str(vars(self))

    def _analysis_type(self) -> str:
        return TYPE_RULE

    @classmethod
    def is_panther_managed(cls) -> bool:
        return cls.__module__.startswith("pypanther.rules")

    @abc.abstractmethod
    def rule(self, event: PantherEvent) -> bool:
        raise NotImplementedError("You must implement the rule method in your rule class.")

    def severity(self, event: PantherEvent) -> SeverityType:
        return self.default_severity

    def title(self, event: PantherEvent) -> str:
        return self.display_name if self.display_name else self.id

    def dedup(self, event: PantherEvent) -> str:
        return self.title(event)

    def destinations(self, event: PantherEvent) -> list[str]:
        return self.default_destinations

    def runbook(self, event: PantherEvent) -> str:
        return self.default_runbook

    def reference(self, event: PantherEvent) -> str:
        return self.default_reference

    def description(self, event: PantherEvent) -> str:
        return self.default_description

    def alert_context(self, event: PantherEvent) -> dict:
        return {}

    def __init_subclass__(cls, **kwargs):
        """
        Creates a copy of all class attributes to avoid mod
        child.tags.append("Foo")
        parent.tags.append("Foo") # not inherited by children of parent
        """
        for attr in RULE_ALL_ATTRS:
            if attr not in cls.__dict__:
                try:
                    v = getattr(cls, attr)
                except AttributeError:
                    v = None

                if v is not None:
                    setattr(cls, attr, copy.deepcopy(v))
        super().__init_subclass__(**kwargs)

    @classmethod
    def asdict(cls):
        """Returns a dictionary representation of the class."""
        return {key: try_asdict(getattr(cls, key)) for key in RULE_ALL_ATTRS if hasattr(cls, key)}

    @classmethod
    def validate_config(cls) -> None:
        """To be defined by subclasses when an out-of-the-box rules requires configuration before use."""

    @classmethod
    def validate(cls, _validate_config: bool = True) -> None:
        """
        Validates this PantherRule.

        Parameters
        ----------
            _validate_config: true if any configuration should be validated, false otherwise. Only meant to be used by Panther.

        """
        RuleAdapter.validate_python(cls.asdict())
        if _validate_config:
            cls.validate_config()

        # instantiation confirms that abstract methods are implemented
        cls()

    @classmethod
    def override(
        cls,
        log_types: Optional[List[str]] = None,
        id: Optional[str] = None,
        create_alert: Optional[bool] = None,
        dedup_period_minutes: Optional[NonNegativeInt] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        scheduled_queries: Optional[List[str]] = None,
        summary_attributes: Optional[List[str]] = None,
        tests: Optional[List[RuleTest]] = None,
        threshold: Optional[PositiveInt] = None,
        tags: Optional[List[str]] = None,
        reports: Optional[Dict[str, List[str]]] = None,
        default_severity: Optional[Severity] = None,
        default_description: Optional[str] = None,
        default_reference: Optional[str] = None,
        default_runbook: Optional[str] = None,
        default_destinations: Optional[List[str]] = None,
    ):
        for key, val in locals().items():
            if key == "cls":
                continue

            if val is not None:
                setattr(cls, key, val)

    @classmethod
    def run_tests(
        cls,
        get_data_model: Callable[[str], Optional[DataModel]],
        _validate_config: bool = True,
    ) -> list[RuleTestResult]:
        """
        Runs all RuleTests in this Rules' Test attribute over this Rule.

        Parameters
        ----------
            get_data_model: a helper function that will return a DataModel given a log type.
            _validate_config: true if tests are being run should validate any configuration, false otherwise. Only meant to be used by Panther.

        Returns
        -------
            a list of RuleTestResult objects.

        """
        cls.validate(_validate_config)
        rule = cls()

        return [rule.run_test(test, get_data_model) for test in rule.tests]

    def run_test(
        self,
        test: RuleTest,
        get_data_model: Callable[[str], Optional[DataModel]],
    ) -> RuleTestResult:
        """
        Runs a unit test over this Rule.

        Parameters
        ----------
            test: the RuleTest to run.
            get_data_model: a helper function that will return a DataModel given a log type.

        Returns
        -------
            a RuleTestResult with the test result. If the Passed attribute is True,
            then this tests passed.

        """
        log = test.log_data()
        log_type = log.get("p_log_type", "default")

        event = PantherEvent(log, get_data_model(log_type))

        patches: list[Any] = []
        for each_mock in test.mocks:
            kwargs = {
                each_mock.object_name: MagicMock(
                    return_value=each_mock.return_value,
                    side_effect=each_mock.side_effect,
                ),
            }
            if each_mock.new is not None:
                kwargs[each_mock.object_name] = each_mock.new
            p = patch.multiple(test._module, **kwargs)
            try:
                p.start()
            except AttributeError:
                p = patch.multiple(self, **kwargs)
                p.start()

        try:
            detection_result = self.run(event, {}, {}, False)

            if (
                detection_result.detection_exception is not None
                or detection_result.detection_output != test.expected_result
            ):
                return RuleTestResult(
                    passed=False,
                    detection_result=detection_result,
                    test=test,
                    rule_id=self.id,
                )

            if isinstance(detection_result.destinations_exception, UnknownDestinationError):
                # ignore unknown destinations during testing
                detection_result.destinations_exception = None

            aux_func_exceptions = {
                "title": detection_result.title_exception,
                "description": detection_result.description_exception,
                "reference": detection_result.reference_exception,
                "severity": detection_result.severity_exception,
                "runbook": detection_result.runbook_exception,
                "destinations": detection_result.destinations_exception,
                "dedup": detection_result.dedup_exception,
                "alert_context": detection_result.alert_context_exception,
            }

            if any(True for _, exc in aux_func_exceptions.items() if exc is not None):
                return RuleTestResult(
                    passed=False,
                    detection_result=detection_result,
                    test=test,
                    rule_id=self.id,
                )

            if any(
                [
                    test.expected_severity is not None and test.expected_severity != detection_result.severity_output,
                    test.expected_title is not None and test.expected_title != detection_result.title_output,
                    test.expected_dedup is not None and test.expected_dedup != detection_result.dedup_output,
                    test.expected_runbook is not None and test.expected_runbook != detection_result.runbook_output,
                    test.expected_reference is not None
                    and test.expected_reference != detection_result.reference_output,
                    test.expected_description is not None
                    and test.expected_description != detection_result.description_output,
                    test.expected_alert_context is not None
                    and test.expected_alert_context != json.loads(detection_result.alert_context_output),
                ],
            ):
                return RuleTestResult(
                    passed=False,
                    detection_result=detection_result,
                    test=test,
                    rule_id=self.id,
                )

        finally:
            for p in patches:
                p.stop()

        return RuleTestResult(
            passed=True,
            detection_result=detection_result,
            test=test,
            rule_id=self.id,
        )

    def run(
        self,
        event: PantherEvent,
        outputs: Dict,
        outputs_names: Dict,
        batch_mode: bool = True,
    ) -> DetectionResult:
        result = DetectionResult(
            detection_id=self.id,
            detection_severity=self.default_severity,
            detection_type=TYPE_RULE,
            # set default to not alert
            trigger_alert=False,
        )

        try:
            result.detection_output = self.rule(event)
            self._require_bool(self.rule.__name__, result.detection_output)
        except Exception as e:
            result.detection_exception = e

        if isinstance(result.detection_output, bool) and result.detection_output:
            result.trigger_alert = True
        if batch_mode and not result.trigger_alert:
            # In batch mode (log analysis), there is no need to run the rest of the functions
            # if the detection isn't going to trigger an alert
            return result

        self.ctx_mgr = noop
        if DISABLE_OUTPUT:
            self.ctx_mgr = suppress_output

        result.title_output, result.title_exception = self._get_title(event)
        result.description_output, result.description_exception = self._get_description(event)
        result.reference_output, result.reference_exception = self._get_reference(event)
        result.severity_output, result.severity_exception = self._get_severity(event)
        result.runbook_output, result.runbook_exception = self._get_runbook(event)
        result.destinations_output, result.destinations_exception = self._get_destinations(
            event,
            outputs,
            outputs_names,
        )
        result.dedup_output, result.dedup_exception = self._get_dedup(event)
        result.alert_context_output, result.alert_context_exception = self._get_alert_context(event)

        if batch_mode:
            # batch mode ignores errors
            # in the panther backend, we check if any error occured during running and if we get one,
            # we return a detection error instead of an alert. To make sure alerts are still returned,
            # we need to set these to None.
            result.title_exception = None
            result.description_exception = None
            result.reference_exception = None
            result.severity_exception = None
            result.runbook_exception = None
            result.destinations_exception = None
            result.dedup_exception = None
            result.alert_context_exception = None

        return result

    def _get_title(self, event: Mapping) -> Tuple[str, Optional[Exception]]:
        try:
            with self.ctx_mgr():
                title = self.title(event)

            self._require_str(self.title.__name__, title)
        except Exception as e:
            title = self.display_name
            if not title or not isinstance(title, str):
                title = self.id
            return title, e

        return truncate(title, MAX_GENERATED_FIELD_SIZE), None

    # Returns the dedup string for this detection match
    def _get_dedup(self, event: Mapping) -> Tuple[Optional[str], Optional[Exception]]:
        e = None
        dedup_string = ""
        try:
            with self.ctx_mgr():
                dedup_string = self.dedup(event)

            self._require_str(self.dedup.__name__, dedup_string)
        except Exception as err:
            e = err

        if dedup_string == "" or not isinstance(dedup_string, str):
            dedup_string, _ = self._get_title(event)
            if dedup_string == "" or not isinstance(dedup_string, str):
                dedup_string = f"defaultDedupString:{self.id}"

        return truncate(dedup_string, MAX_DEDUP_STRING_SIZE), e

    def _get_description(
        self,
        event: Mapping,
    ) -> Tuple[str, Optional[Exception]]:
        try:
            with self.ctx_mgr():
                description = self.description(event)

            self._require_str(self.description.__name__, description)
        except Exception as e:
            return "", e

        return truncate(description, MAX_GENERATED_FIELD_SIZE), None

    def _get_reference(self, event: Mapping) -> Tuple[str, Optional[Exception]]:
        try:
            with self.ctx_mgr():
                reference = self.reference(event)

            self._require_str(self.reference.__name__, reference)
        except Exception as e:
            return "", e

        return truncate(reference, MAX_GENERATED_FIELD_SIZE), None

    def _get_runbook(self, event: Mapping) -> Tuple[Optional[str], Optional[Exception]]:
        try:
            with self.ctx_mgr():
                runbook = self.runbook(event)

            self._require_str(self.runbook.__name__, runbook)
        except Exception as e:
            return "", e

        return truncate(runbook, MAX_GENERATED_FIELD_SIZE), None

    def _get_severity(self, event: Mapping) -> Tuple[Optional[str], Optional[Exception]]:
        try:
            with self.ctx_mgr():
                severity: str = self.severity(event)

            self._require_str(self.severity.__name__, severity)
            severity = severity.upper()
            if severity == SEVERITY_DEFAULT:
                return self.default_severity, None
            if severity not in SEVERITY_TYPES:
                raise AssertionError(
                    f"Expected severity to be any of the following: [{SEVERITY_TYPES!s}], got [{severity}] instead.",
                )
        except Exception as e:
            return self.default_severity, e

        return severity, None

    def _get_alert_context(self, event: Mapping) -> Tuple[Optional[str], Optional[Exception]]:
        try:
            with self.ctx_mgr():
                alert_context = self.alert_context(event)

            self._require_mapping(self.alert_context.__name__, alert_context)
            serialized_alert_context = json.dumps(alert_context, default=PantherEvent.json_encoder)
        except Exception as err:
            return json.dumps({ALERT_CONTEXT_ERROR_KEY: repr(err)}), err

        if len(serialized_alert_context) > MAX_ALERT_CONTEXT_SIZE:
            # If context exceeds max size, return empty one
            alert_context_error = (
                f"alert_context size is [{len(serialized_alert_context)}] characters,"
                f" bigger than maximum of [{MAX_ALERT_CONTEXT_SIZE}] characters"
            )
            return json.dumps({ALERT_CONTEXT_ERROR_KEY: alert_context_error}), None

        return serialized_alert_context, None

    def _get_destinations(
        self,
        event: Mapping,
        outputs: Dict,
        outputs_display_names: Dict,
    ) -> Tuple[Optional[List[str]], Optional[Exception]]:
        try:
            with self.ctx_mgr():
                destinations = self.destinations(event)
            self._require_str_list(self.destinations.__name__, destinations)
        except Exception as e:
            return None, e

        # Return early if destinations returned None
        if destinations is None:
            return None, None

        # Return early if destinations is an empty list (alert dest. suppression)
        if len(destinations) == 0:
            return ["SKIP"], None

        # Check for (in)valid destinations
        invalid_destinations = []
        standardized_destinations: List[str] = []

        # Standardize the destinations
        for each_destination in destinations:
            # case for valid display name
            if (
                each_destination in outputs_display_names
                and outputs_display_names[each_destination].destination_id not in standardized_destinations
            ):
                standardized_destinations.append(outputs_display_names[each_destination].destination_id)
            # case for valid UUIDv4
            elif each_destination in outputs and each_destination not in standardized_destinations:
                standardized_destinations.append(each_destination)
            else:
                invalid_destinations.append(each_destination)

        if len(standardized_destinations) > MAX_DESTINATIONS_SIZE:
            # If generated field exceeds max size, truncate it
            standardized_destinations = standardized_destinations[:MAX_DESTINATIONS_SIZE]

        if invalid_destinations:
            try:
                # raise to get a stack trace
                raise UnknownDestinationError("Invalid Destinations", invalid_destinations)
            except UnknownDestinationError as e:
                return standardized_destinations, e

        return standardized_destinations, None

    def _require_bool(self, method_name: str, value: Any):
        return self._require_scalar(method_name, bool, value)

    def _require_str(self, method_name: str, value: Any):
        return self._require_scalar(method_name, str, value)

    def _require_mapping(self, method_name: str, value: Any):
        return self._require_scalar(method_name, Mapping, value)

    def _require_scalar(self, method_name: str, typ: Type, value: Any):
        if not isinstance(value, typ):
            raise FunctionReturnTypeError(
                f"detection [{self.id}] method [{method_name}] returned [{type(value).__name__}], expected [{typ.__name__}]",
            )

    def _require_str_list(self, method_name: str, value: Any):
        if value is None:
            return
        if not isinstance(value, list) or not all(isinstance(x, (str, bool)) for x in value):
            raise FunctionReturnTypeError(
                f"detection [{self.id}] method [{method_name}] returned [{type(value).__name__}], expected a list",
            )


@contextlib.contextmanager
def suppress_output():
    original_stdout = sys.stdout
    original_stderr = sys.stderr

    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stdout = original_stdout
        sys.stderr = original_stderr


@contextlib.contextmanager
def noop():
    yield


def panther_managed(cls: Type[Rule]) -> Type[Rule]:
    """Decorator to apply to OOTB rules written by Panther."""
    cls._tests = cls.tests  # type: ignore
    cls.tests = []
    return cls



from pypanther import LogType, Rule, RuleTest, Severity, panther_managed
from pypanther.helpers.panther_base_helpers import github_alert_context

github_repository_transfer_tests: list[RuleTest] = [
    RuleTest(
        name="Public Repo Created",
        expected_result=False,
        log={
            "_document_id": "abCD",
            "action": "repo.create",
            "actor": "example-actor",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-11 22:40:20.268",
            "created_at": "2022-12-11 22:40:20.268",
            "org": "example-io",
            "repo": "example-io/oops",
            "visibility": "public",
        },
    ),
    RuleTest(
        name="Repo Transfer Outgoing",
        expected_result=True,
        log={
            "_document_id": "BodJtQIrT3kWMIQpm1ANew",
            "action": "repo.transfer_outgoing",
            "actor": "user-name",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-14 19:16:31.299",
            "created_at": "2022-12-14 19:16:31.299",
            "org": "your-organization",
            "repo": "your-organizatoin/project_repo",
            "visibility": "private",
        },
    ),
    RuleTest(
        name="Repo Transfer Start",
        expected_result=True,
        log={
            "_document_id": "BodJtQIrT3kWMIQpm1ANew",
            "action": "repo.transfer_start",
            "actor": "user-name",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-14 19:16:31.299",
            "created_at": "2022-12-14 19:16:31.299",
            "org": "your-organization",
            "repo": "your-organizatoin/project_repo",
            "visibility": "private",
        },
    ),
    RuleTest(
        name="Repository Transfer",
        expected_result=True,
        log={
            "_document_id": "CFyS8UJsQjJfCgsmTLI6mQ",
            "action": "repo.transfer",
            "actor": "org-user",
            "actor_location": {"country_code": "US"},
            "at_sign_timestamp": "2022-12-14 19:21:01.035",
            "created_at": "2022-12-14 19:21:01.035",
            "org": "your-organization",
            "repo": "your-organization/project_repo",
            "visibility": "private",
        },
    ),
]


@panther_managed
class GithubRepositoryTransfer(Rule):
    default_description = "A user accepted a request to receive a transferred Github repository, a  Github repository was transferred to another repository network, or a user sent a request to transfer a repository to another user or organization."
    display_name = "Github Repository Transfer"
    default_reference = "https://docs.github.com/en/enterprise-server@3.3/repositories/creating-and-managing-repositories/transferring-a-repository\n\nhttps://docs.github.com/en/enterprise-cloud@latest/admin/monitoring-activity-in-your-enterprise/reviewing-audit-logs-for-your-enterprise/audit-log-events-for-your-enterprise#repo-category-actions"
    default_runbook = "Please check with the referenced users or their supervisors to ensure the transferring of this repository is expected and allowed."
    default_severity = Severity.MEDIUM
    tags = ["Github Repository", "Github Repository Transfer", "Repository", "Transfer"]
    log_types = [LogType.GITHUB_AUDIT]
    id = "Github.Repository.Transfer-prototype"
    summary_attributes = ["action"]
    tests = github_repository_transfer_tests

    def rule(self, event):
        # Return True to match the log event and trigger an alert.
        return event.get("action", "") in ("repo.transfer", "repo.transfer_outgoing", "repo.transfer_start")

    def title(self, event):
        # (Optional) Return a string which will be shown as the alert title.
        # If no 'dedup' function is defined, the return value of this method
        # will act as deduplication string.
        action = event.get("action", "")
        if action == "repo.transfer":
            # return something like: A user accepted a request to receive a transferred repository.
            return f"Github User [{event.get('actor', 'NO_ACTOR_FOUND')}] accepted a request to receive repository [{event.get('repo', 'NO_REPO_NAME_FOUND')}] in [{event.get('org', 'NO_ORG_NAME_FOUND')}]."
        if action == "repo.transfer_outgoing":
            # return something like: A repository was transferred to another repository network.
            return f"Github User [{event.get('actor', 'NO_ACTOR_FOUND')}] transferred repository [{event.get('repo', 'NO_REPO_NAME_FOUND')}] in [{event.get('org', 'NO_ORG_NAME_FOUND')}]."
        if action == "repo.transfer_start":
            # return something like: A user sent a request to transfer a
            # repository to another user or organization.
            return f"Github User [{event.get('actor', 'NO_ACTOR_FOUND')}] sent a request to transfer repository [{event.get('repo', 'NO_REPO_NAME_FOUND')}] to another user or organization."
        return ""

    def alert_context(self, event):
        #  (Optional) Return a dictionary with additional data to be included in the alert
        # sent to the SNS/SQS/Webhook destination
        return github_alert_context(event)
